import numpy as np

from libertem.udf import UDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        # the last tile contains only two frames:
        assert tile.shape[0] in (7, 2)

        assert len(tile.shape) == 3
        assert tile.shape[1:] == (16, 16)
        assert self.results.pixelsum.shape == (tile.shape[0],)
        axes = tuple(range(1, len(tile.shape)))
        self.results.pixelsum[:] += np.sum(tile, axis=axes)


def test_sum_tiles(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(7, 16, 16),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(2, 3)))


class PixelsumForCropped(UDF):
    """
    like above but different asserts in process_tile
    """
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_tile(self, tile):
        # the last tile contains only two frames:
        assert tile.shape[0] in (7, 2)
        assert len(tile.shape) == 3
        assert self.results.pixelsum.shape == (tile.shape[0],)
        axes = tuple(range(1, len(tile.shape)))
        self.results.pixelsum[:] += np.sum(tile, axis=axes)


def test_mem_cropped(lt_ctx):
    """
    to make sure the memory dataset works fine with cropping:
    """
    data = _mk_random(size=(16, 16, 24, 24), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(7, 7, 7),
                            num_partitions=2, sig_dims=2)

    buf = np.zeros((256, 24, 24), dtype="float32")
    for p in dataset.get_partitions():
        for tile in p.get_tiles():
            assert tuple(tile.tile_slice.shape)[0] in (7, 2)
            assert tuple(tile.tile_slice.shape)[1:] in [(7, 7),
                                                        (7, 3),
                                                        (3, 7),
                                                        (3, 3)]
            buf[tile.tile_slice.get()] = tile.data

    assert np.allclose(
        buf.reshape(data.shape), data
    )


def test_cropped(lt_ctx):
    data = _mk_random(size=(16, 16, 24, 24), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(7, 7, 7),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumForCropped()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(2, 3)))


def test_roi_extra_dimension_shape(lt_ctx):
    """
    Test sum over the pixels for 2-dimensional dataset

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them

    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    class ExtraShapeUDF(UDF):
        def get_result_buffers(self):
            return {
                'test': self.buffer(
                    kind="nav", extra_shape=(2,), dtype="float32"
                ),
                'test2': self.buffer(
                    kind="sig", extra_shape=(2,), dtype="float32"
                ),
                # 'test3': self.buffer(
                #     kind="single", extra_shape=(2,), dtype="float32"
                # )
            }

        def process_tile(self, tile):
            tile_slice = self.meta.slice
            self.results.test[:] = np.ones(tuple(tile_slice.shape.nav)) * (1, 2)
            framecount = np.prod(tuple(tile_slice.shape.nav))
            self.results.test2[:] += np.ones(tuple(tile_slice.shape.sig) + (2, )) * framecount
            # self.results.test3[:] += (framecount, 2*framecount)

        def merge(self, dest, src):
            dest.test[:] = src.test[:]
            dest.test2[:] += src.test2[:]
            # dest.test3[:] += src.test3[:]

    extra = ExtraShapeUDF()
    roi = _mk_random(size=dataset.shape.nav, dtype=bool)
    res = lt_ctx.run_udf(dataset=dataset, udf=extra, roi=roi)

    navcount = np.count_nonzero(roi)

    print(data.shape, res['test'].data.shape)
    assert res['test'].data.shape == tuple(dataset.shape.nav) + (2,)
    assert res['test2'].data.shape == tuple(dataset.shape.sig) + (2,)
    # assert res['test3'].data.shape == (2,)
    assert np.allclose(res['test'].raw_data, (1, 2))
    assert np.allclose(res['test2'].raw_data, navcount)
    # assert np.allclose(res['test3'].raw_data, (navcount, 2*navcount))



import pathlib
import numpy as np
import pytest

from libertem.udf.sumsigudf import SumSigUDF


@pytest.mark.parametrize(
    "sync_offset", [-2, 0, 2]
)
def test_write_npy(tmpdir_factory, raw_data_8x8x8x8_path, lt_ctx_fast, sync_offset):
    write_dir = tmpdir_factory.mktemp('data')
    write_path = pathlib.Path(write_dir) / 'write_raw_data.npy'
    ds = lt_ctx_fast.load(
        'raw',
        raw_data_8x8x8x8_path,
        np.float32,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
        sync_offset=sync_offset,
    )
    lt_ctx_fast.export_dataset(ds, path=write_path)

    res = lt_ctx_fast.run_udf(udf=SumSigUDF(), dataset=ds)
    sum_ds = res['intensity'].data

    sum_file = np.load(write_path).sum(axis=(2, 3))
    assert np.allclose(sum_ds, sum_file)


def test_bad_extension(tmpdir_factory, raw_data_8x8x8x8_path, lt_ctx_fast):
    write_dir = tmpdir_factory.mktemp('data')
    write_path = pathlib.Path(write_dir) / 'bad.foo'
    ds = lt_ctx_fast.load(
        'raw',
        raw_data_8x8x8x8_path,
        np.float32,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
    )
    with pytest.raises(ValueError):
        lt_ctx_fast.export_dataset(ds, path=write_path)


def test_no_overwrite(tmpdir_factory, raw_data_8x8x8x8_path, lt_ctx_fast):
    write_dir = tmpdir_factory.mktemp('data')
    write_path = pathlib.Path(write_dir) / 'exists.npy'

    some_data = np.ones((5, 5))
    np.save(write_path, some_data)

    ds = lt_ctx_fast.load(
        'raw',
        raw_data_8x8x8x8_path,
        np.float32,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
    )
    with pytest.raises(FileExistsError):
        lt_ctx_fast.export_dataset(ds, path=write_path)



import numpy as np
import numba
import sparse

from libertem.common.math import prod
from libertem.common.sparse import is_sparse
from libertem.common.numba import (
    numba_ravel_multi_index_multi, numba_unravel_index_multi
)


@numba.njit(cache=True, nogil=True)
def _correct_numba_inplace(buffer, dark_image, gain_map, exclude_pixels, repair_environments,
        repair_counts):
    '''
    Numerical work horse to perform detector corrections

    This function uses blocked processing for cache efficiency, hence the nested loops. It is
    about 4x faster than a naive numpy implementation.

    Parameters
    ----------

    buffer:
        (n, m) with data, modified in-place.

    dark_image:
        (m) with dark frame to be subtracted first

    gain_map:
        (m) with gain map to multiply the data with after subtraction

    exclude_pixels:
        int(k) array of indices in the flattened signal dimension to patch

    repair_environments:
        Array with environments for each pixel to use as a reference.
        Array of shape (k, max_repair_counts)

    repair_counts:
        Array(int) of length k with number of valid entries in each entry in repair_environments.


    Returns
    -------
    buffer (modified in-place)
    '''
    nav_blocksize = 4
    sig_blocksize = 2**20 // (512 * 4 * nav_blocksize)

    nav_blocks = buffer.shape[0] // nav_blocksize
    nav_remainder = buffer.shape[0] % nav_blocksize

    sig_blocks = buffer.shape[1] // sig_blocksize
    sig_remainder = buffer.shape[1] % sig_blocksize

    def _get_dark_px(sig):
        if dark_image is None:
            return 0
        return dark_image[sig]

    def _get_gain_px(sig):
        if gain_map is None:
            return 1
        return gain_map[sig]

    for nav_block in range(nav_blocks):
        # Dark and gain blocked in sig
        for sig_block in range(sig_blocks):
            for nav in range(nav_block * nav_blocksize, (nav_block + 1) * nav_blocksize):
                for sig in range(sig_block * sig_blocksize, (sig_block + 1) * sig_blocksize):
                    buffer[nav, sig] = (buffer[nav, sig] - _get_dark_px(sig)) * _get_gain_px(sig)

        # Dark and gain remainder of sig blocks
        for nav in range(nav_block * nav_blocksize, (nav_block + 1) * nav_blocksize):
            for sig in range(sig_blocks*sig_blocksize, sig_blocks*sig_blocksize + sig_remainder):
                buffer[nav, sig] = (buffer[nav, sig] - _get_dark_px(sig)) * _get_gain_px(sig)

        # Hole repair blocked in nav
        for i, p in enumerate(exclude_pixels):
            if repair_counts[i] > 0:  # Avoid div0
                for nav in range(nav_block * nav_blocksize, (nav_block + 1) * nav_blocksize):
                    acc = 0
                    for index in repair_environments[i, :repair_counts[i]]:
                        acc += buffer[nav, index]
                    buffer[nav, p] = acc / repair_counts[i]

    # Processing unblocked nav remainder
    for nav in range(nav_blocks * nav_blocksize, nav_blocks * nav_blocksize + nav_remainder):
        # Dark and gain unblocked in sig
        for sig in range(buffer.shape[1]):
            buffer[nav, sig] = (buffer[nav, sig] - _get_dark_px(sig)) * _get_gain_px(sig)

        # Hole repair
        for i, p in enumerate(exclude_pixels):
            if repair_counts[i] > 0:  # Avoid div0
                acc = 0
                for index in repair_environments[i, :repair_counts[i]]:
                    acc += buffer[nav, index]
                buffer[nav, p] = acc / repair_counts[i]

    return buffer


@numba.njit(cache=True, nogil=True)
def environments(excluded_pixels, sigshape):
    '''
    Calculate a hypercube surface around a pixel, excluding frame boundaries

    Returns
    -------
    repairs, repair_counts
        repairs : numpy.ndarray
            Array with shape (exclude_pixels, sig_dims, indices)
        repair_counts : numpy.ndarray
            Array with length exclude_pixels, containing the number of pixels
            in the repair environment
    '''

    max_repair_count = 3**len(sigshape) - 1
    num_pixels = len(excluded_pixels[0])
    repairs = np.zeros((num_pixels, len(sigshape), max_repair_count), dtype=np.intp)
    repair_counts = np.zeros(num_pixels, dtype=np.intp)
    all_indices = np.arange(3**len(sigshape), dtype=np.intp)
    coord_shape = np.full(len(sigshape), 3, dtype=np.intp)
    coord_offsets = numba_unravel_index_multi(all_indices, coord_shape) - 1
    for i in range(num_pixels):
        repair_count = 0
        for position in range(coord_offsets.shape[1]):
            select = False
            for dim in range(coord_offsets.shape[0]):
                coord = coord_offsets[dim, position] + excluded_pixels[dim, i]
                # Any of the coordinates is different
                select += (coord != excluded_pixels[dim, i])
            for dim in range(coord_offsets.shape[0]):
                coord = coord_offsets[dim, position] + excluded_pixels[dim, i]
                # All of the coordinates are within bounds
                select *= (coord >= 0)
                select *= (coord < sigshape[dim])
            if select:
                for dim in range(coord_offsets.shape[0]):
                    coord = coord_offsets[dim, position] + excluded_pixels[dim, i]
                    repairs[i, dim, repair_count] = coord
                repair_count += 1
        repair_counts[i] = repair_count

    return repairs, repair_counts


class RepairValueError(ValueError):
    pass


@numba.njit(cache=True, nogil=True)
def flatten_filter(excluded_pixels, repairs, repair_counts, sig_shape):
    '''
    Flatten excluded pixels and repair environments and filter for collisions

    Ravel indices to flattened signal dimension and
    removed damaged pixels from all repair environments, i.e. only use
    "good" pixels.
    '''
    excluded_flat = numba_ravel_multi_index_multi(excluded_pixels, sig_shape)

    max_repair_count = 3**len(sig_shape) - 1
    new_repair_counts = np.zeros_like(repair_counts)
    repair_flat = np.zeros((len(excluded_flat), max_repair_count), dtype=np.intp)

    excluded_dict = {}
    for i in excluded_flat:
        excluded_dict[i] = True

    for i in range(len(excluded_flat)):
        a = numba_ravel_multi_index_multi(repairs[i, ..., :repair_counts[i]], sig_shape)
        nonzero_index = 0
        for j in range(repair_counts[i]):
            if a[j] not in excluded_dict:
                repair_flat[i, nonzero_index] = a[j]
                nonzero_index += 1
        new_repair_counts[i] = nonzero_index
        if new_repair_counts[i] == 0:
            pass
            # TODO fix for Numba
            # raise RepairValueError("Repair environment for pixel %i is empty" % i)

    return (excluded_flat, repair_flat, new_repair_counts)


def correct(
        buffer, dark_image=None, gain_map=None, excluded_pixels=None, repair_descriptor=None,
        inplace=False, sig_shape=None, allow_empty=False):
    '''
    Function to perform detector corrections

    This function delegates the processing to a function written with numba that is
    about 4x faster than a naive numpy implementation.

    Parameters
    ----------

    buffer:
        shape (*nav, *sig) with data. It is modified in-place if inplace==True.

    dark_image:
        shape (*sig) with dark frame to be subtracted first

    gain_map:
        shape (*sig) with gain map to multiply the data with after subtraction

    exclude_pixels:
        int(sigs, k) array of indices in the signal dimension to patch.
        The first dimension is the number of signal dimensions, the second the number of pixels

    repair_descriptor : RepairDescriptor
        This allows to re-use the calculation and filtering of repair environments when
        specified instead of exclude_pixels. This is particularly advantageous for tiled processing.

    inplace:
        If True, modify the input buffer in-place.
        If False (default), copy the input buffer before correcting.

    Returns
    -------
    shape (*nav, *sig) If inplace==True, this is :code:`buffer` modified in-place.
    '''
    s = buffer.shape

    if dark_image is not None:
        sig_shape = dark_image.shape
        dark_image = dark_image.flatten()
    if gain_map is not None:
        sig_shape = gain_map.shape
        gain_map = gain_map.flatten()
    if sig_shape is None:
        raise ValueError("need either `dark_image`, `gain_map`, or `sig_shape`")
    nav_shape = s[0:-len(sig_shape)]

    if inplace:
        if buffer.dtype.kind not in ('f', 'c'):
            raise TypeError("In-place correction only supported for floating point data.")
        out = buffer
    else:
        # astype() is always a copy even if it is the same dtype
        out = buffer.astype(np.result_type(np.float32, buffer))

    if repair_descriptor is None:
        repair_descriptor = RepairDescriptor(
            sig_shape=sig_shape,
            excluded_pixels=excluded_pixels,
            allow_empty=allow_empty
        )
    else:
        repair_descriptor.check_empty_repairs(allow_empty=allow_empty)
        if excluded_pixels is not None:
            raise ValueError("Invalid arguments: Bot repair_descriptor and excluded_pixels set")

    _correct_numba_inplace(
        buffer=out.reshape((prod(nav_shape), prod(sig_shape))),
        dark_image=dark_image,
        gain_map=gain_map,
        exclude_pixels=repair_descriptor.exclude_flat,
        repair_environments=repair_descriptor.repair_flat,
        repair_counts=repair_descriptor.repair_counts,
    )
    return out


class RepairDescriptor:
    def __init__(self, sig_shape, excluded_pixels=None, allow_empty=False):
        if excluded_pixels is None:
            excluded_pixels = np.zeros((len(sig_shape), 0), dtype=np.intp)
        else:
            excluded_pixels = np.array(excluded_pixels)

        repairs, repair_counts = environments(excluded_pixels, np.array(sig_shape))

        self.exclude_flat, self.repair_flat, self.repair_counts = flatten_filter(
            excluded_pixels, repairs, repair_counts, sig_shape
        )
        self.check_empty_repairs(allow_empty=allow_empty)

    def empty_repairs(self):
        return np.argwhere(self.repair_counts == 0)

    def check_empty_repairs(self, allow_empty):
        if not allow_empty:
            empty = self.empty_repairs()
            if len(empty) > 0:
                raise RepairValueError(
                    f"Empty repair environments for pixel(s) number {empty}."
                )


def correct_dot_masks(masks, gain_map, excluded_pixels=None, allow_empty=False):
    mask_shape = masks.shape
    sig_shape = gain_map.shape
    masks = masks.reshape((-1, prod(sig_shape)))

    if excluded_pixels is not None:
        if is_sparse(masks):
            result = sparse.DOK(masks)
        else:
            result = masks.copy()
        desc = RepairDescriptor(sig_shape, excluded_pixels=excluded_pixels, allow_empty=allow_empty)
        for e, r, c in zip(desc.exclude_flat, desc.repair_flat, desc.repair_counts):
            result[:, e] = 0
            rep = masks[:, e] / c
            # We have to loop because of sparse.pydata limitations
            for m in range(result.shape[0]):
                for rr in r[:c]:
                    result[m, rr] = result[m, rr] + rep[m]
        if is_sparse(result):
            result = sparse.COO(result)
    else:
        result = masks
    result = result * gain_map.flatten()
    return result.reshape(mask_shape)



import tornado
import logging
from .notebook_generator.notebook_generator import notebook_generator
from .notebook_generator.copy import copy_notebook

from .state import SharedState

log = logging.getLogger(__name__)


class DownloadScriptHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def get(self, compoundUuid: str):
        compoundAnalysis = self.state.compound_analysis_state[compoundUuid]
        analysis_ids = compoundAnalysis['details']['analyses']
        ds_id = self.state.analysis_state[analysis_ids[0]]['dataset']
        ds = self.state.dataset_state.datasets[ds_id]
        dataset = {
            "type": ds["params"]["params"]["type"],
            "params": ds['converted']
        }
        main_type = compoundAnalysis['details']['mainType'].lower()
        ds_name = ds["params"]["params"]["name"]

        analysis_details = []
        for id in analysis_ids:
            analysis_details.append(self.state.analysis_state[id]['details'])
        conn = self.state.executor_state.get_cluster_params()
        buf = notebook_generator(conn, dataset, analysis_details)
        self.set_header('Content-Type', 'application/vnd.jupyter.cells')
        self.set_header(
            'Content-Disposition',
            f'attachment; filename="{main_type}_{ds_name}.ipynb"',
        )
        self.write(buf.getvalue())


class CopyScriptHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def get(self, compoundUuid: str):
        compoundAnalysis = self.state.compound_analysis_state[compoundUuid]
        analysis_ids = compoundAnalysis['details']['analyses']
        ds_id = self.state.analysis_state[analysis_ids[0]]['dataset']
        ds = self.state.dataset_state.datasets[ds_id]
        dataset = {
            "type": ds["params"]["params"]["type"],
            "params": ds['converted']
        }

        analysis_details = []
        for id in analysis_ids:
            analysis_details.append(self.state.analysis_state[id]['details'])
        conn = self.state.executor_state.get_cluster_params()
        notebook = copy_notebook(conn, dataset, analysis_details)
        self.write(notebook)



import pathlib
import numpy as np
import pytest

from libertem.contrib.convert_transposed import convert_dm4_transposed, _convert_transposed_ds

from utils import _mk_random


@pytest.mark.parametrize(
    'nav_shape',
    [
        (6,),
        (3, 4),
    ]
)
@pytest.mark.parametrize(
    'sig_shape',
    [
        (8,),
        (7, 5),
    ]
)
def test_functional(lt_ctx, nav_shape, sig_shape, tmpdir_factory):
    nav_dims = len(nav_shape)
    sig_dims = len(sig_shape)

    dir = pathlib.Path(tmpdir_factory.mktemp('convert_transposed'))
    data = _mk_random(size=sig_shape + nav_shape, dtype='float32')
    # Transpose to put the nav dims before the sig dims, but retaining
    # the order of the dimensions within each group
    # e.g. (0, 1, 2, 3) => (2, 3, 0, 1)
    converted_data = np.moveaxis(
        data,
        tuple(range(data.ndim)),
        tuple((*range(nav_dims, data.ndim), *range(nav_dims))),
    )
    ds = lt_ctx.load(
        'memory',
        data=data,
        # sig_dims=nav_dims because nav_dims are at the end!
        # needed so that we correctly partition the array
        sig_dims=nav_dims,
        num_partitions=2,
    )

    convert_path = dir / 'out.npy'
    _convert_transposed_ds(lt_ctx, ds, convert_path)

    # Now that we are transposed sig_dims=sig_dims
    ds_npy = lt_ctx.load('npy', path=convert_path, sig_dims=sig_dims)
    assert ds_npy.shape.to_tuple() == converted_data.shape
    assert np.allclose(converted_data, np.load(convert_path))


def test_both_args_raise(lt_ctx):
    with pytest.raises(ValueError):
        convert_dm4_transposed(
            'tata.dm4',
            'out.npy',
            ctx=lt_ctx,
            num_cpus=42,
        )


# Tests on actual DM4 datasets are in tests/io/datasets/test_dm_single.py
# to use the dm4 file fixtures defined there



import numpy as np
import pytest

from libertem.io.dataset.base import TilingScheme, DataSetException
from libertem.common import Shape
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.stddev import StdDevUDF
from libertem.io.dataset.memory import MemoryDataSet
from libertem.utils.devices import detect
from sparseconverter import (
    CUPY_SCIPY_CSC, NUMPY, SCIPY_CSR, SPARSE_COO, SPARSE_GCXS,
    CUPY, get_backend, get_device_class
)
from libertem.udf.base import UDFProtocol

from utils import _mk_random, dataset_correction_verification, set_device_class, roi_as_sparse


d = detect()
has_cupy = d['cudas'] and d['has_cupy']


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(UDFProtocol.BACKEND_ALL)
)
def test_get_macrotile(backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16))
        ds = MemoryDataSet(
            data=data,
            tileshape=(16, 16, 16),
            num_partitions=2,
            array_backends=(backend, ) if backend is not None else None,
        )

        p = next(ds.get_partitions())
        mt = p.get_macrotile()
        assert tuple(mt.shape) == (128, 16, 16)

        if backend is None:
            assert get_backend(data) == get_backend(mt.data)
        else:
            get_backend(mt.data) == backend


@pytest.mark.parametrize(
    'backend', [None, NUMPY, SPARSE_COO, CUPY, CUPY_SCIPY_CSC]
)
@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_correction(lt_ctx, with_roi, backend, as_sparse):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16))
        ds = MemoryDataSet(
            data=data,
            tileshape=(16, 16, 16),
            num_partitions=2,
            array_backends=(backend, ) if backend is not None else None,
        )

        if with_roi:
            roi = np.zeros(ds.shape.nav, dtype=bool)
            roi[:1] = True
            if as_sparse:
                roi = roi_as_sparse(roi)
        else:
            roi = None

        dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@pytest.mark.parametrize(
    'backend', [None, SPARSE_GCXS, CUPY_SCIPY_CSC]
)
def test_positive_sync_offset(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        udf = SumSigUDF()
        data = _mk_random(size=(8, 8, 8, 8))
        sync_offset = 2

        ds_with_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=sync_offset,
            array_backends=(backend, ) if backend is not None else None
        )

        p0 = next(ds_with_offset.get_partitions())
        assert p0._start_frame == 2
        assert p0.slice.origin == (0, 0, 0)

        tileshape = Shape(
            (2,) + tuple(ds_with_offset.shape.sig),
            sig_dims=ds_with_offset.shape.sig.dims
        )
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=tileshape,
            dataset_shape=ds_with_offset.shape,
        )

        for p in ds_with_offset.get_partitions():
            for t in p.get_tiles(tiling_scheme=tiling_scheme, array_backend=backend):
                if backend is None:
                    assert get_backend(t.data) == get_backend(data)
                else:
                    assert get_backend(t.data) == backend

        assert p.slice.origin == (48, 0, 0)
        assert p.slice.shape[0] == 16

        ds_with_no_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=0,
            array_backends=(backend, ) if backend is not None else None
        )
        result = lt_ctx.run_udf(dataset=ds_with_no_offset, udf=udf)
        result = result['intensity'].raw_data[sync_offset:]

        result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
        result_with_offset = result_with_offset['intensity'].raw_data[
            :ds_with_offset._meta.image_count - sync_offset
        ]

        assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    'backend', [NUMPY, SCIPY_CSR, CUPY]
)
def test_negative_sync_offset(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        udf = SumSigUDF()
        data = _mk_random(size=(8, 8, 8, 8))
        sync_offset = -2

        ds_with_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=sync_offset,
            array_backends=(backend, ) if backend is not None else None
        )

        p0 = next(ds_with_offset.get_partitions())
        assert p0._start_frame == -2
        assert p0.slice.origin == (0, 0, 0)

        tileshape = Shape(
            (2,) + tuple(ds_with_offset.shape.sig),
            sig_dims=ds_with_offset.shape.sig.dims
        )
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=tileshape,
            dataset_shape=ds_with_offset.shape,
        )

        for p in ds_with_offset.get_partitions():
            for t in p.get_tiles(tiling_scheme=tiling_scheme, array_backend=backend):
                if backend is None:
                    assert get_backend(t.data) == get_backend(data)
                else:
                    assert get_backend(t.data) == backend

        assert p.slice.origin == (48, 0, 0)
        assert p.slice.shape[0] == 16

        ds_with_no_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=0,
            array_backends=(backend, ) if backend is not None else None
        )
        result = lt_ctx.run_udf(dataset=ds_with_no_offset, udf=udf)
        result = result['intensity'].raw_data[
            :ds_with_no_offset._meta.image_count - abs(sync_offset)
        ]

        result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
        result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

        assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
@pytest.mark.parametrize(
    'backend', [None, SPARSE_GCXS, CUPY_SCIPY_CSC]
)
def test_positive_sync_offset_with_roi(lt_ctx, backend, as_sparse):
    with set_device_class(get_device_class(backend)):
        roi = np.random.choice([False], (8, 8))
        roi[0:1] = True
        if as_sparse:
            roi = roi_as_sparse(roi)

        udf = SumSigUDF()

        data = np.random.randn(8, 8, 8, 8).astype("float32")
        ds = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=0,
            array_backends=(backend, ) if backend is not None else None
        )
        result = lt_ctx.run_udf(dataset=ds, udf=udf)
        result = result['intensity'].raw_data

        sync_offset = 2

        ds_with_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=sync_offset,
            array_backends=(backend, ) if backend is not None else None
        )
        result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
        result_with_offset = result_with_offset['intensity'].raw_data
        roi = np.random.choice([False], (8, 8))
        roi[0:1] = True

        result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
        result_with_offset = result_with_offset['intensity'].raw_data

        assert np.allclose(result[sync_offset:8 + sync_offset], result_with_offset)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
@pytest.mark.parametrize(
    'backend', [None, SCIPY_CSR, CUPY]
)
def test_negative_sync_offset_with_roi(lt_ctx, backend, as_sparse):
    with set_device_class(get_device_class(backend)):
        udf = SumSigUDF()

        data = np.random.randn(8, 8, 8, 8).astype("float32")
        ds = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=0,
            array_backends=(backend, ) if backend is not None else None
        )
        result = lt_ctx.run_udf(dataset=ds, udf=udf)
        result = result['intensity'].raw_data

        sync_offset = -2

        ds_with_offset = MemoryDataSet(
            data=data,
            tileshape=(2, 8, 8),
            num_partitions=4,
            sync_offset=sync_offset,
            array_backends=(backend, ) if backend is not None else None
        )

        roi = np.random.choice([False], (8, 8))
        roi[0:1] = True
        if as_sparse:
            roi = roi_as_sparse(roi)

        result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
        result_with_offset = result_with_offset['intensity'].raw_data

        assert np.allclose(result[:8 + sync_offset], result_with_offset[abs(sync_offset):])


def test_scheme_too_large():
    data = _mk_random(size=(16, 16, 16, 16))
    ds = MemoryDataSet(
        data=data,
        # tileshape=(16, 16, 16),
        num_partitions=2,
    )

    partitions = ds.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme, array_backend=NUMPY)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + ds.shape.sig)


@pytest.mark.parametrize(
    'nav_shape', (None, (3, 7), (14*8 + 1, ), (13, 19, 23))
)
@pytest.mark.parametrize(
    'sig_shape', (None, (19, 23), (19 * 23, ))
)
@pytest.mark.parametrize(
    'sync_offset', (0, -3, 7, 13*7, -14*8)
)
@pytest.mark.parametrize(
    'sig_dims', (None, 1, 2)
)
def test_sig_nav_dims_sync(nav_shape, sig_shape, sync_offset, sig_dims, prime_raw_data, lt_ctx):
    if sig_shape is None and sig_dims is None and nav_shape is not None and len(nav_shape) < 2:
        pytest.xfail("This one doesn't reshape nav as the test requires.")
    print('nav_shape', nav_shape)
    print('sig_shape', sig_shape)
    print('sync_offset', sync_offset)
    print('sig_dims', sig_dims)
    throws_value_error = (
        sig_shape is not None and sig_dims is not None and len(sig_shape) != sig_dims
    )
    if nav_shape is not None:
        expected_nav_shape = nav_shape
    elif sig_dims is not None:
        expected_nav_shape = prime_raw_data.shape[:-sig_dims]
    elif sig_shape is not None:
        expected_nav_shape = prime_raw_data.shape[:-len(sig_shape)]
    else:
        expected_nav_shape = prime_raw_data.shape[:-2]

    if sig_shape is not None:
        expected_sig_shape = sig_shape
    elif sig_dims is not None:
        expected_sig_shape = prime_raw_data.shape[-sig_dims:]
    elif nav_shape is not None:
        expected_sig_shape = prime_raw_data.shape[len(nav_shape):]
    else:
        expected_sig_shape = prime_raw_data.shape[-2:]

    if throws_value_error:
        with pytest.raises(ValueError):
            ds = lt_ctx.load(
                'memory',
                data=prime_raw_data,
                nav_shape=nav_shape,
                sig_shape=sig_shape,
                sync_offset=sync_offset,
                sig_dims=sig_dims
            )
    else:
        ds = lt_ctx.load(
            'memory',
            data=prime_raw_data,
            nav_shape=nav_shape,
            sig_shape=sig_shape,
            sync_offset=sync_offset,
            sig_dims=sig_dims
        )
        assert tuple(ds.shape.sig) == expected_sig_shape
        assert tuple(ds.shape.nav) == expected_nav_shape

        # FIXME also check correctness of results
        udf = StdDevUDF()
        lt_ctx.run_udf(dataset=ds, udf=udf)


def test_exception_no_datashape(lt_ctx_fast):
    with pytest.raises(DataSetException):
        lt_ctx_fast.load('memory', tileshape=(5, 6, 7))



import inspect

from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
from .helper import GeneratorHelper
import libertem.udf.FEM as FEM


class FEMTemplate(GeneratorHelper):

    short_name = "fem"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return ["from libertem.udf.FEM import FEMUDF"]

    def get_docs(self):
        title = "FEM Analysis"
        docs_rst = inspect.getdoc(FEM.FEMUDF)
        docs = self.format_docs(title, docs_rst)
        return docs

    def convert_params(self):
        cx = self.params['cx']
        cy = self.params['cy']
        ri = self.params['ri']
        ro = self.params['ro']
        params = f"center=({cx},{cy}), rad_in={ri}, rad_out={ro}"
        return params

    def get_analysis(self):
        params = self.convert_params()
        temp_analysis = [
                    f"fem_udf = FEMUDF({params})",
                    "fem_result = ctx.run_udf(dataset=ds, udf=fem_udf)",
        ]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = [
            "plt.figure()",
            "plt.imshow(fem_result['intensity'])",
            "plt.colorbar()",
        ]
        return ['\n'.join(plot)]


class FEMAnalysis(BaseAnalysis, id_="FEM"):
    TYPE = "UDF"

    def get_udf(self):
        center = (self.parameters["cy"], self.parameters["cx"])
        rad_in = self.parameters["ri"]
        rad_out = self.parameters["ro"]
        return FEM.FEMUDF(center=center, rad_in=rad_in, rad_out=rad_out)

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        return AnalysisResultSet([
            AnalysisResult(raw_data=udf_results['intensity'].data,
                           visualized=visualize_simple(
                               udf_results['intensity'].data.reshape(self.dataset.shape.nav),
                               damage=damage
                            ),
                           key="intensity", title="intensity",
                           desc="result from SD calculation over ring"),
        ])

    @classmethod
    def get_template_helper(cls):
        return FEMTemplate



import copy

from libertem.analysis.base import Analysis
from .template import TemplateBase


class Literal:
    def __init__(self, value):
        self.value = value

    def __repr__(self) -> str:
        return self.value


class CodeTemplate(TemplateBase):
    """
    Generate code using template from :class:`TemplateBase` and parameters
    from GUI.

    new code segments can be generated by adding templates in :class:`TemplateBase`
    and combining parameters accordingly.

    code for dataset, dependencies, connection are generated here. Analysis
    specific code are handled in `libertem.analysis.helper`.
    """

    def __init__(self, connection, dataset, compound_analysis):
        self.conn = connection['connection']
        self.ds = dataset
        self.compound_analysis = compound_analysis

        self.analysis_helper = {}
        for analysis in self.compound_analysis:

            type = analysis['analysisType']
            params = analysis['parameters']
            cls = Analysis.get_analysis_by_type(type)
            helperCls = cls.get_template_helper()
            helper = helperCls(params)
            self.analysis_helper[type] = helper

    def dataset(self):
        ds_type = self.ds['type']
        ds_params = copy.copy(self.ds['params'])
        data = {
            'type': ds_type,
            'params': ds_params,
        }
        result = []
        # because the io_backend parameter is not trivially serializable
        # like the others, we special case it here:
        if "io_backend" in ds_params:
            ds_params["io_backend"] = Literal("io_backend")
            data['backend_cls'] = self.ds["params"]["io_backend"].__class__.__name__
            result.append(
                self.code_formatter(self.format_template(self.temp_ds_backend, data))
            )
        result.append(
            self.code_formatter(self.format_template(self.temp_ds, data))
        )
        return "\n".join(result)

    def dependency(self):
        """
        dependencies required for generated script.

        Common dependencies are available in :class:`TemplateBase`.
        Any additional dependencies can be added through
        corresponding analysis helper.
        """
        extra_dep = []

        if self.conn['type'].lower() == "tcp":
            extra_dep.extend(self.temp_dep_conn)

        if "io_backend" in self.ds["params"]:
            data = {
                "backend_cls": self.ds["params"]["io_backend"].__class__.__name__
            }
            extra_dep.append(
                self.code_formatter(self.format_template(self.temp_dep_ds, data))
            )

        for helper in self.analysis_helper.values():
            analysis_dep = helper.get_dependency()
            if analysis_dep is not None:
                extra_dep.extend(analysis_dep)
        dep = self.temp_dep + extra_dep
        return self.code_formatter('\n'.join(dep))

    def initial_setup(self):
        return "%matplotlib nbagg"

    def connection(self):
        docs = ["# Connection"]
        if self.conn['type'].lower() == "tcp":
            link = "https://libertem.github.io/LiberTEM/usage.html#starting-a-custom-cluster"
            more_info = f"[For more info]({link})"
            docs.append(f"Connecting to dask cluster, {more_info}")
            data = {'conn_url': self.conn['address']}
            ctx = self.format_template(self.temp_conn, data)
            docs = '\n'.join(docs)
            return ctx, docs
        elif self.conn['type'].lower() == "local":
            docs.append("This starts a local cluster that is accessible through ctx.")
            ctx = "ctx = lt.Context()"
            docs = '\n'.join(docs)
            return ctx, docs
        else:
            raise ValueError("unknown connection type")

    def analysis(self):
        form_analysis = []

        for helper in self.analysis_helper.values():

            plot_ = list(map(self.code_formatter, helper.get_plot()))
            analy_ = self.code_formatter(helper.get_analysis())
            docs_ = self.code_formatter(helper.get_docs())
            save_ = self.code_formatter(helper.get_save())

            form_analysis.append((docs_, analy_, plot_, save_))

        return form_analysis



import functools
import asyncio
from typing import Callable, TYPE_CHECKING, TypeVar

from contextlib import asynccontextmanager

from libertem.common.async_utils import (
    adjust_event_loop_policy, sync_to_async, async_generator_eager
)
from libertem.common.executor import JobExecutor, AsyncJobExecutor
from libertem.common.tracing import TracedThreadPoolExecutor


if TYPE_CHECKING:
    from libertem.udf.base import UDFRunner

T = TypeVar('T')


class ResourceError(RuntimeError):
    """
    Thrown when there is a resource mismatch, for example if the task requests
    resources that are not available in the worker pool.
    """
    pass


class BaseJobExecutor(JobExecutor):
    def get_udf_runner(self) -> type['UDFRunner']:
        from libertem.udf.base import UDFRunner
        return UDFRunner

    def ensure_async(self, pool=None):
        """
        Returns an asynchronous executor; by default just wrap into `AsyncAdapter`.
        """
        return AsyncAdapter(wrapped=self, pool=pool)


class AsyncAdapter(AsyncJobExecutor):
    """
    Wrap a synchronous JobExecutor and allow to use it as AsyncJobExecutor. All methods are
    converted to async and executed in a separate thread.
    """
    def __init__(self, wrapped, pool=None):
        self._wrapped = wrapped
        if pool is None:
            pool = AsyncAdapter.make_pool()
        self._pool = pool

    @classmethod
    def make_pool(cls):
        pool = TracedThreadPoolExecutor(1)
        pool.submit(adjust_event_loop_policy).result()
        pool.submit(lambda: asyncio.set_event_loop(asyncio.new_event_loop())).result()
        return pool

    def ensure_sync(self):
        return self._wrapped

    @asynccontextmanager
    async def scatter(self, obj):
        try:
            res = await sync_to_async(self._wrapped.scatter.__enter__, self._pool)
            yield res
        finally:
            exit_fn = functools.partial(
                self._wrapped.scatter.__exit__,
                None, None, None,  # FIXME: exc_type, exc_value, traceback?
            )
            await sync_to_async(exit_fn, self._pool)

    async def run_tasks(self, tasks, params_handle, cancel_id):
        """
        run a number of Tasks
        """
        gen = self._wrapped.run_tasks(tasks, params_handle, cancel_id)
        agen = async_generator_eager(gen, self._pool)
        async for i in agen:
            yield i

    async def run_function(self, fn: Callable[..., T], *args, **kwargs) -> T:
        """
        run a callable :code:`fn` on an arbitrary worker node
        """
        fn_with_args = functools.partial(self._wrapped.run_function, fn, *args, **kwargs)
        return await sync_to_async(fn_with_args, self._pool)

    async def run_each_partition(self, partitions, fn, all_nodes=False):
        fn_with_args = functools.partial(
            self._wrapped.run_each_partition, partitions, fn, all_nodes
        )
        return await sync_to_async(fn_with_args, self._pool)

    async def map(self, fn, iterable):
        """
        Run a callable :code:`fn` for each item in iterable, on arbitrary worker nodes

        Parameters
        ----------

        fn : callable
            Function to call. Should accept exactly one parameter.

        iterable : Iterable
            Which elements to call the function on.
        """
        fn_with_args = functools.partial(
            self._wrapped.map, fn, iterable,
        )
        return await sync_to_async(fn_with_args, self._pool)

    async def run_each_host(self, fn, *args, **kwargs):
        fn_with_args = functools.partial(self._wrapped.run_each_host, fn, *args, **kwargs)
        return await sync_to_async(fn_with_args, self._pool)

    async def run_each_worker(self, fn, *args, **kwargs):
        fn_with_args = functools.partial(self._wrapped.run_each_worker, fn, *args, **kwargs)
        return await sync_to_async(fn_with_args, self._pool)

    async def close(self):
        """
        Cleanup resources used by this executor, if any, including the wrapped executor.
        """
        res = await sync_to_async(self._wrapped.close, self._pool)
        if self._pool:
            self._pool.shutdown()
        return res

    async def cancel(self, cancel_id):
        """
        cancel execution identified by cancel_id
        """
        return await sync_to_async(
            functools.partial(self._wrapped.cancel, cancel_id=cancel_id),
            self._pool
        )

    async def get_available_workers(self):
        return await sync_to_async(self._wrapped.get_available_workers)

    async def get_resource_details(self):
        return await sync_to_async(self._wrapped.get_resource_details)

    def get_udf_runner(self) -> type['UDFRunner']:
        from libertem.udf.base import UDFRunner
        return UDFRunner



import typing
from typing import Optional
from collections.abc import Generator, Sequence

import numpy as np
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common.shape import Shape
from libertem.common.math import prod
from libertem.io.utils import get_partition_shape
from libertem.io.dataset.base import DataSetException, MMapBackend
from libertem.common.messageconverter import MessageConverter
from libertem.io.corrections.corrset import CorrectionSet
from .partition import BasePartition, Partition

if typing.TYPE_CHECKING:
    from libertem.common.executor import JobExecutor, TaskCommHandler
    from libertem.io.dataset.base import IOBackend, Decoder, DataSetMeta
    from numpy import typing as nt


class DataSet:
    # The default partition size in bytes
    MAX_PARTITION_SIZE = 512*1024*1024

    def __init__(self, io_backend: Optional["IOBackend"] = None):
        self._cores = 1
        self._sync_offset: Optional[int] = 0
        self._sync_offset_info = None
        self._image_count = 0
        self._nav_shape_product = 0
        self._io_backend = io_backend
        self._meta: Optional[DataSetMeta] = None

    def initialize(self, executor) -> "DataSet":
        """
        Perform possibly expensive initialization, like pre-loading metadata.

        This is run on the master node, but can execute parts on workers, for example
        if they need to access the data stored on worker nodes, using the passed executor
        instance.

        If you need the executor around for later operations, for example when creating
        the partitioning, save a reference here!

        Should return the possibly modified `DataSet` instance (if a method running
        on a worker is changing `self`, these changes won't automatically be transferred back
        to the master node)
        """
        raise NotImplementedError()

    def set_num_cores(self, cores: int) -> None:
        self._cores = cores

    def get_sync_offset_info(self):
        """
        Check sync_offset specified and returns number of frames skipped and inserted
        """
        if not -1*self._image_count < self._sync_offset < self._image_count:
            raise DataSetException(
                "sync_offset should be in (%s, %s), which is (-image_count, image_count)"
                % (-1*self._image_count, self._image_count)
            )
        return {
            "frames_skipped_start": max(0, self._sync_offset),
            "frames_ignored_end": max(
                0, self._image_count - self._nav_shape_product - self._sync_offset
            ),
            "frames_inserted_start": abs(min(0, self._sync_offset)),
            "frames_inserted_end": max(
                0, self._nav_shape_product - self._image_count + self._sync_offset
            )
        }

    def get_num_partitions(self) -> int:
        """
        Returns the number of partitions the dataset should be split into.

        The default implementation sizes partition such that they
        fit into 512MB of float data in memory, regardless of their
        native dtype. At least :code:`self._cores` partitions
        are created.
        """
        partition_size_float_px = self.MAX_PARTITION_SIZE // 4
        dataset_size_px = prod(self.shape)
        num: int = max(self._cores, dataset_size_px // partition_size_float_px)
        return num

    def get_slices(self):
        """
        Return the partition slices for the dataset
        """
        return BasePartition.make_slices(
            shape=self.shape,
            num_partitions=self.get_num_partitions(),
            sync_offset=self._sync_offset,
        )

    def get_partitions(self) -> Generator[Partition, None, None]:
        """
        Return a generator over all Partitions in this DataSet. Should only
        be called on the master node.
        """
        raise NotImplementedError()

    @property
    def dtype(self) -> "nt.DTypeLike":
        """
        The "native" data type
        (either one matching the data on disk, or one that is closest)
        """
        raise NotImplementedError()

    @property
    def shape(self) -> Shape:
        """
        The shape of the DataSet, as it makes sense for the application domain
        (for example, 4D for pixelated STEM)
        """
        raise NotImplementedError()

    @property
    def array_backends(self) -> Sequence[ArrayBackend]:
        """
        The array backends the dataset can return data as.

        Defaults to only NumPy arrays

        .. versionadded:: 0.11.0
        """
        return (NUMPY, CUDA)

    def check_valid(self) -> bool:
        """
        check validity of the DataSet. this will be executed (after initialize) on a worker node.
        should raise DataSetException in case of errors, return True otherwise.
        """
        raise NotImplementedError()

    @classmethod
    def detect_params(cls, path: str, executor: "JobExecutor"):
        """
        Guess if path can be opened using this DataSet implementation and
        detect parameters.

        returns dict of detected parameters if path matches this dataset type,
        returns False if path is most likely not of a matching type.
        """
        # FIXME: return hints for the user and additional values,
        # for example number of signal elements
        raise NotImplementedError()

    @classmethod
    def get_msg_converter(cls) -> type[MessageConverter]:
        raise NotImplementedError()

    @property
    def diagnostics(self):
        """
        Diagnostics common for all DataSet implementations
        """
        p = next(self.get_partitions())

        return self.get_diagnostics() + [
            {"name": "Partition shape",
             "value": str(p.shape)},
            {"name": "Number of partitions",
             "value": str(len(list(self.get_partitions())))},
            {"name": "Number of frames skipped at the beginning",
             "value": self._sync_offset_info["frames_skipped_start"]},
            {"name": "Number of frames ignored at the end",
            "value": self._sync_offset_info["frames_ignored_end"]},
            {"name": "Number of blank frames inserted at the beginning",
            "value": self._sync_offset_info["frames_inserted_start"]},
            {"name": "Number of blank frames inserted at the end",
            "value": self._sync_offset_info["frames_inserted_end"]}
        ]

    def get_diagnostics(self):
        """
        Get relevant diagnostics for this dataset, as a list of
        dicts with keys name, value, where value may be string or
        a list of dicts itself. Subclasses should override this method.
        """
        return []

    def partition_shape(
        self,
        dtype: "nt.DTypeLike",
        target_size: int,
        min_num_partitions: Optional[int] = None,
        containing_shape: Optional[Shape] = None,
    ) -> tuple[int, ...]:
        """
        Calculate partition shape for the given ``target_size``

        Parameters
        ----------
        dtype : numpy.dtype or str
            data type of the dataset

        target_size : int
            target size in bytes - how large should each partition be?

        min_num_partitions : int
            minimum number of partitions desired. Defaults to the number of workers in the cluster.

        Returns
        -------
        Tuple[int, ...]
            the shape calculated from the given parameters
        """
        if min_num_partitions is None:
            min_num_partitions = self._cores
        if containing_shape is None:
            containing_shape = self.shape
        return get_partition_shape(
            dataset_shape=containing_shape,
            target_size_items=target_size // np.dtype(dtype).itemsize,
            min_num=min_num_partitions
        )

    @classmethod
    def get_supported_extensions(cls) -> set[str]:
        """
        Return supported extensions as a set of strings.

        Plain extensions only, no pattern!
        """
        return set()

    def get_cache_key(self) -> str:
        raise NotImplementedError()

    @classmethod
    def get_default_io_backend(cls) -> "IOBackend":
        import platform
        if platform.system() == "Windows":
            from libertem.io.dataset.base import BufferedBackend
            return BufferedBackend()
        return MMapBackend()

    @classmethod
    def get_supported_io_backends(cls) -> list[str]:
        """
        Get the supported I/O backends as list of their IDs. Some DataSet
        implementations with a custom backend may return an empty list here.
        """
        return ["mmap", "buffered", "direct"]

    def get_io_backend(self) -> "IOBackend":
        if self._io_backend is None:
            return self.get_default_io_backend()
        return self._io_backend

    def get_correction_data(self) -> CorrectionSet:
        """
        Correction parameters that are part of this DataSet.
        This should only be called after the DataSet is initialized.

        Returns
        -------
        CorrectionSet
            correction parameters that are part of this DataSet
        """
        return CorrectionSet()

    def supports_correction(self):
        return True

    def get_decoder(self) -> Optional["Decoder"]:
        return None

    def get_base_shape(self, roi: Optional[np.ndarray]) -> tuple[int, ...]:
        return (1,) + (1,) * (self.shape.sig.dims - 1) + (self.shape.sig[-1],)

    def adjust_tileshape(
        self, tileshape: tuple[int, ...], roi: Optional[np.ndarray]
    ) -> tuple[int, ...]:
        """
        Final veto of the DataSet in the tileshape negotiation process,
        make sure that corrections are taken into account!
        """
        return tileshape

    def need_decode(
        self,
        read_dtype: "nt.DTypeLike",
        roi: Optional[np.ndarray],
        corrections: Optional[CorrectionSet],
    ) -> bool:
        io_backend = self.get_io_backend().get_impl()
        return io_backend.need_copy(
            decoder=self.get_decoder(),
            roi=roi,
            native_dtype=self.meta.raw_dtype,
            read_dtype=read_dtype,
            sync_offset=self._sync_offset,
            corrections=corrections,
        )

    def get_min_sig_size(self) -> int:
        """
        minimum signal size, in number of elements
        """
        return 4 * 4096 // np.dtype(self.meta.raw_dtype).itemsize

    def get_max_io_size(self) -> Optional[int]:
        """
        Override this method to implement a custom maximum I/O size (in bytes)
        """
        return None

    @property
    def meta(self) -> Optional["DataSetMeta"]:
        return self._meta

    def get_task_comm_handler(self) -> "TaskCommHandler":
        from libertem.common.executor import NoopCommHandler
        return NoopCommHandler()


class WritableDataSet:
    pass



import inspect
from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
import libertem.udf.stddev as std
from libertem.analysis.getroi import get_roi
from .helper import GeneratorHelper


class SDTemplate(GeneratorHelper):

    short_name = "sd"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return [
            "from matplotlib import colors",
            "from libertem import masks",
            "from libertem.udf.stddev import StdDevUDF"
            ]

    def get_docs(self):
        title = "SD Analysis"
        docs_rst = inspect.getdoc(std.StdDevUDF)
        docs = self.format_docs(title, docs_rst)
        return docs

    def get_analysis(self):
        roi = self.get_roi()
        temp_analysis = [
                    f"{roi}",
                    "sd_udf  = StdDevUDF()",
                    "sd_result = ctx.run_udf(dataset=ds, roi=roi, udf=sd_udf)",
                    ]
        return '\n'.join(temp_analysis)

    # FIXME  `plt.colorbar()` creates error while testing
    # moreinfo  https://github.com/LiberTEM/LiberTEM/pull/801#pullrequestreview-453173083
    def get_plot(self):
        plot = [
            "plt.figure()",
            "plt.imshow(sd_result['varsum'].data, norm=colors.LogNorm())",
        ]
        return ['\n'.join(plot)]

    def get_save(self):
        save = "np.save('sd_result.npy', sd_result['varsum'])"
        return save


class SDAnalysis(BaseAnalysis, id_="SD_FRAMES"):
    TYPE = 'UDF'

    def get_udf(self):
        return std.StdDevUDF()

    def get_roi(self):
        return get_roi(params=self.parameters, shape=self.dataset.shape.nav)

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        return AnalysisResultSet([
            AnalysisResult(
                raw_data=udf_results['std'].data,
                visualized=visualize_simple(
                    udf_results['std'].data, logarithmic=True,
                    damage=True
                ),
                key="intensity", title="intensity [log]",
                desc="Standard deviation of frames log-scaled"
            ),
            AnalysisResult(
                raw_data=udf_results['std'],
                visualized=visualize_simple(
                    udf_results['std'].data, logarithmic=False,
                    damage=True
                ),
                key="intensity_lin",
                title="intensity [lin]",
                desc="Standard deviation of frames lin-scaled"
            ),
        ])

    @classmethod
    def get_template_helper(cls):
        return SDTemplate



import os
import stat

from libertem.common.messageconverter import MessageConverter  # NOQA: F401
from libertem.common.progress import ProgressState


class Message:
    """
    possible messages - the translation of our python datatypes to json types
    """

    def initial_state(self, jobs, datasets, analyses, compound_analyses):
        return {
            "status": "ok",
            "messageType": "INITIAL_STATE",
            "datasets": datasets,
            "jobs": jobs,
            "analyses": analyses,
            "compoundAnalyses": compound_analyses,
        }

    def cluster_conn_error(self, msg):
        return {
            "status": "error",
            "messageType": "CLUSTER_CONN_ERROR",
            "msg": msg,
        }

    def snooze(self, msg):
        return {
            "status": "ok",
            "messageType": "SNOOZE",
            "msg": msg,
        }

    def unsnooze(self, msg):
        return {
            "status": "ok",
            "messageType": "UNSNOOZE",
            "msg": msg,
        }

    def unsnooze_done(self, msg):
        return {
            "status": "ok",
            "messageType": "UNSNOOZE_DONE",
            "msg": msg,
        }

    def config(self, config):
        return {
            "status": "ok",
            "messageType": "CONFIG",
            "config": config,
        }

    def cluster_details(self, details):
        return {
            "status": "ok",
            "messageType": "CLUSTER_DETAILS",
            "details": details,
        }

    def create_dataset(self, dataset, details):
        return {
            "status": "ok",
            "messageType": "CREATE_DATASET",
            "dataset": dataset,
            "details": details,
        }

    def create_dataset_error(self, dataset, msg):
        return {
            "status": "error",
            "messageType": "CREATE_DATASET_ERROR",
            "dataset": dataset,
            "msg": msg,
        }

    def delete_dataset(self, dataset):
        return {
            "status": "ok",
            "messageType": "DELETE_DATASET",
            "dataset": dataset,
        }

    def dataset_detect(self, params, info=None):
        return {
            "status": "ok",
            "messageType": "DATASET_DETECTED",
            "datasetParams": params,
            "datasetInfo": info,
        }

    def dataset_detect_failed(self, path,
                              reason="could not automatically determine dataset format"):
        return {
            "status": "error",
            "messageType": "DATASET_DETECTION_FAILED",
            "path": path,
            "msg": reason,
        }

    def start_job(self, serialized_job, analysis_id):
        return {
            "status": "ok",
            "messageType": "JOB_STARTED",
            "job": serialized_job["id"],
            "analysis": analysis_id,
            "details": serialized_job,
        }

    def job_error(self, job_id, msg):
        return {
            "status": "error",
            "messageType": "JOB_ERROR",
            "job": job_id,
            "msg": msg,
        }

    def job_progress(self, job_id: str, state: ProgressState, event: str):
        return {
            "status": "ok",
            "messageType": "JOB_PROGRESS",
            "job": job_id,
            "details": {
                "event": event,
                "numFrames": state.num_frames_total,
                "numFramesComplete": state.num_frames_complete,
            }
        }

    def finish_job(self, serialized_job, num_images, image_descriptions):
        return {
            "status": "ok",
            "messageType": "FINISH_JOB",
            "job": serialized_job["id"],
            "details": serialized_job,
            "followup": {
                "numMessages": num_images,
                "descriptions": image_descriptions,
            },
        }

    def cancel_job(self, job_id):
        return {
            "status": "ok",
            "messageType": "CANCEL_JOB",
            "job": job_id,
        }

    def cancel_done(self, job_id):
        return {
            "status": "ok",
            "messageType": "CANCEL_JOB_DONE",
            "job": job_id,
        }

    def cancel_failed(self, job_id):
        return {
            "status": "error",
            "messageType": "CANCEL_JOB_FAILED",
            "job": job_id,
        }

    def task_result(self, job_id, num_images, image_descriptions):
        return {
            "status": "ok",
            "messageType": "TASK_RESULT",
            "job": job_id,
            "followup": {
                "numMessages": num_images,
                "descriptions": image_descriptions,
            },
        }

    def directory_listing(self, path, files, dirs, drives, places):
        def _details(item):
            return {
                "name":  item["name"],
                "size":  item["stat"].st_size,
                "ctime": item["stat"].st_ctime,
                "mtime": item["stat"].st_mtime,
                "owner": item["owner"],
            }

        return {
            "status": "ok",
            "messageType": "DIRECTORY_LISTING",
            "drives": drives,
            "places": places,
            "path": path,
            "files": [
                _details(f)
                for f in files
            ],
            "dirs": [
                _details(d)
                for d in dirs
            ],
        }

    def browse_stat_result(
        self,
        path: str,
        dirname: str,
        basename: str,
        stat_result: os.stat_result,
    ):
        return {
            "status": "ok",
            "messageType": "STAT_RESULT",
            "path": path,
            "dirname": dirname,
            "basename": basename,
            "stat": {
                "size":  stat_result.st_size,
                "ctime": stat_result.st_ctime,
                "mtime": stat_result.st_mtime,
                "isdir": bool(stat.S_ISDIR(stat_result.st_mode)),
                "isreg": bool(stat.S_ISREG(stat_result.st_mode)),
            }
        }

    def stat_failed(self, path, code, msg, alternative=None):
        return {
            "status": "error",
            "messageType": "STAT_FAILED",
            "path": path,
            "code": code,
            "msg": msg,
            "alternative": alternative,
        }

    def browse_failed(self, path, code, msg, alternative=None):
        return {
            "status": "error",
            "messageType": "DIRECTORY_LISTING_FAILED",
            "path": path,
            "code": code,
            "msg": msg,
            "alternative": alternative,
        }

    def create_analysis(self, uuid, dataset_uuid, analysis_type, parameters):
        return {
            "status": "ok",
            "messageType": "ANALYSIS_CREATED",
            "analysis": uuid,
            "dataset": dataset_uuid,
            "details": {
                "analysisType": analysis_type,
                "parameters": parameters,
            }
        }

    def update_analysis(self, uuid, dataset_uuid, analysis_type, parameters):
        return {
            "status": "ok",
            "messageType": "ANALYSIS_UPDATED",
            "analysis": uuid,
            "dataset": dataset_uuid,
            "details": {
                "analysisType": analysis_type,
                "parameters": parameters,
            }
        }

    def analysis_removed(self, uuid):
        return {
            "status": "ok",
            "messageType": "ANALYSIS_REMOVED",
            "analysis": uuid,
        }

    def analysis_removal_failed(self, uuid, msg):
        return {
            "status": "error",
            "messageType": "ANALYSIS_REMOVAL_FAILED",
            "analysis": uuid,
            "msg": msg,
        }

    def compound_analysis_created(self, serialized):
        msg = {
            "status": "ok",
            "messageType": "COMPOUND_ANALYSIS_CREATED",
        }
        msg.update(serialized)
        return msg

    def compound_analysis_updated(self, serialized):
        msg = {
            "status": "ok",
            "messageType": "COMPOUND_ANALYSIS_UPDATED",
        }
        msg.update(serialized)
        return msg

    def compound_analysis_removed(self, uuid):
        msg = {
            "status": "ok",
            "messageType": "COMPOUND_ANALYSIS_REMOVED",
            "compoundAnalysis": uuid,
        }
        return msg



import numpy as np
import pytest

from libertem.io.dataset import get_extensions
from libertem.io.dataset.base import (
    Partition, _roi_to_nd_indices, DataSetException
)
from libertem.common import Shape, Slice
from libertem.io.dataset.memory import MemoryDataSet
from libertem.io.dataset.base import TilingScheme

from utils import _mk_random


def test_sweep_stackheight():
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data.astype("<u2"),
        num_partitions=2,
    )
    for stackheight in range(1, 256):
        tileshape = Shape(
            (stackheight,) + tuple(dataset.shape.sig),
            sig_dims=dataset.shape.sig.dims
        )
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=tileshape,
            dataset_shape=dataset.shape,
        )
        print("testing with stackheight", stackheight)
        for p in dataset.get_partitions():
            for tile in p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32"):
                pass


def test_num_part_larger_than_num_frames():
    shape = Shape((1, 1, 256, 256), sig_dims=2)
    slice_iter = Partition.make_slices(shape=shape, num_partitions=2)
    next(slice_iter)
    with pytest.raises(StopIteration):
        next(slice_iter)


@pytest.mark.parametrize(
    'repeat', range(30)
)
def test_make_slices(repeat):
    sync_offset = np.random.randint(-5, 5)
    num_part = np.random.randint(1, 41)
    num_frames = np.random.randint(num_part, 333)
    shape = Shape((num_frames,), 0)
    slices = tuple(Partition.make_slices(shape=shape,
                                         num_partitions=num_part,
                                         sync_offset=sync_offset))
    assert len(slices) == num_part
    assert slices[0][0].origin[0] == 0
    assert slices[-1][0].origin[0] + slices[-1][0].shape[0] == num_frames
    assert slices[0][1] == 0 + sync_offset
    assert slices[-1][-1] == num_frames + sync_offset
    assert all(s0[-1] == s1[1] for s0, s1 in zip(slices[:-1], slices[1:]))


def test_roi_to_nd_indices():
    roi = np.full((5, 5), False)
    roi[1, 2] = True
    roi[2, 1:4] = True
    roi[3, 2] = True

    part_slice = Slice(
        origin=(2, 0, 0, 0),
        shape=Shape((2, 5, 16, 16), sig_dims=2)
    )

    assert list(_roi_to_nd_indices(roi, part_slice)) == [
        (2, 1), (2, 2), (2, 3),
                (3, 2)
    ]

    part_slice = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((5, 5, 16, 16), sig_dims=2)
    )

    assert list(_roi_to_nd_indices(roi, part_slice)) == [
                (1, 2),
        (2, 1), (2, 2), (2, 3),         # NOQA: E131
                (3, 2)
    ]


def test_get_extensions():
    exts = get_extensions()
    assert len(exts) >= 15
    assert "mib" in exts
    assert "gtg" in exts
    # etc...


def test_filetype_auto(hdf5, lt_ctx):
    ds = lt_ctx.load("auto", path=hdf5.filename)
    assert ds.ds_path == "data"


def test_filetype_auto_fail_no_path(lt_ctx):
    with pytest.raises(TypeError):
        lt_ctx.load("auto")


def test_filetype_auto_fail_file_does_not_exist(lt_ctx):
    with pytest.raises(DataSetException) as e:
        lt_ctx.load("auto", path="/does/not/exist/believe_me")
    assert e.match("could not determine DataSet type for file")



"""
Types and objects used in the web API
"""

from typing import TYPE_CHECKING, NamedTuple, Optional

from typing_extensions import TypedDict


if TYPE_CHECKING:
    from libertem.analysis.base import AnalysisResultSet
    from libertem.udf.base import UDFResults
    from libertem.io.dataset.base import DataSet


class JobInfo(TypedDict):
    id: str
    analysis: str
    dataset: str


class SerializedJobInfo(TypedDict):
    id: str
    analysis: str


AnalysisParameters = dict


class AnalysisDetails(TypedDict):
    analysisType: str
    parameters: AnalysisParameters


class AnalysisInfo(TypedDict):
    dataset: str
    analysis: str
    jobs: list[str]
    details: AnalysisDetails


class AnalysisResultInfo(NamedTuple):
    details: AnalysisDetails
    results: "AnalysisResultSet"
    job_id: str
    udf_results: Optional["UDFResults"]


class CompAnalysisDetails(TypedDict):
    mainType: str
    analyses: list[str]


class CompoundAnalysisInfo(TypedDict):
    dataset: str
    compoundAnalysis: str
    details: CompAnalysisDetails


class DatasetInfo(TypedDict):
    dataset: "DataSet"
    params: dict
    converted: dict



import queue
from typing import (
    Callable, Optional, Any, TYPE_CHECKING,
    TypeVar,
)
from collections.abc import Generator, Iterable
from contextlib import contextmanager
import multiprocessing as mp

import cloudpickle
from opentelemetry import trace
import numpy as np
from typing_extensions import Protocol, Literal

from libertem.common.scheduler import WorkerSet
from libertem.common.threading import set_num_threads, mitigations
from libertem.io.dataset.base import Partition


if TYPE_CHECKING:
    from libertem.udf.base import UDFParams, UDFRunner
    from opentelemetry.trace import SpanContext

ResourceDef = dict[
    Literal[
        'CPU', 'compute', 'ndarray', 'CUDA',
    ],
    int
]

tracer = trace.get_tracer(__name__)


class ExecutorError(Exception):
    pass


class JobCancelledError(Exception):
    """
    raised by async executors in run_tasks() or run_each_partition() if the task was cancelled
    """
    pass


class Environment:
    '''
    Create the environment to run a task, in particular thread count.
    '''
    def __init__(
        self,
        threads_per_worker: Optional[int],
        threaded_executor: bool,
        worker_context: Optional["WorkerContext"] = None,
    ):
        self._threads_per_worker = threads_per_worker
        self._threaded_executor = threaded_executor
        self._worker_context = worker_context

    @property
    def threads_per_worker(self) -> Optional[int]:
        """
        int or None : number of threads that a UDF is allowed to use in the `process_*` method.
                      For numba, pyfftw, OMP, MKL, OpenBLAS, this limit is set automatically;
                      this property can be used for other cases, like manually creating
                      thread pools.
                      None means no limit is set, and the UDF can use any number of threads
                      it deems necessary (should be limited to system limits, of course).

        See also: :func:`libertem.common.threading.set_num_threads`

        .. versionadded:: 0.7.0
        """
        return self._threads_per_worker

    @property
    def threaded_executor(self) -> bool:
        """
        Whether or not the executor uses threading for parallelism.
        If this flag is set, mitigations for common threading issues are
        automatically applied.
        """
        return self._threaded_executor

    @property
    def worker_context(self) -> Optional["WorkerContext"]:
        """
        A :code:`WorkerContext` instance, if available, as supplied
        by the `JobExecutor`. This is used to manage streaming communication
        between the main process and the workers. If :code:`None` is
        returned, streaming communication with workers is not available.
        """
        return self._worker_context

    @contextmanager
    def enter(self):
        """
        Note: we are using the @contextmanager decorator here,
        because with separate `__enter__`, `__exit__` methods,
        we can't easily delegate to `set_num_threads`, or other
        contextmanagers that may come later.
        """
        with set_num_threads(self._threads_per_worker):
            if self.threaded_executor:
                with mitigations():
                    yield self
            else:
                yield self


class TaskProtocol(Protocol):
    '''
    Interface for tasks
    '''
    def __call__(self, params: "UDFParams", env: Environment):
        pass

    def get_tracing_span_context(self) -> "SpanContext":
        ...

    def get_partition(self) -> Partition:
        ...

    def get_resources(self) -> ResourceDef:
        ...


T = TypeVar('T')
V = TypeVar('V')


class JobExecutor:
    '''
    Interface to execute functions on workers.
    '''
    def run_function(self, fn: Callable[..., T], *args, **kwargs) -> T:
        """
        run a callable :code:`fn` on any worker
        """
        raise NotImplementedError()

    def scatter_update(self, handle, obj):
        '''
        Update :code:`handle` to point to :code:`obj`

        Must have been scattered before using :meth:`scatter`.

        Parameters
        ----------
        handle
            The handle, as returned from :meth:`scatter`.

        obj
            Some kind of Python object. Must be serializable.
        '''
        raise NotImplementedError()

    def scatter_update_patch(self, handle, patch):
        '''
        Update :code:`handle` by remotely calling :code:`obj.patch(patch)` on
        the underlying object.

        The referenced object must have a :code:`patch` method, and
        must have been scattered before using :meth:`scatter`.

        Parameters
        ----------
        handle
            The handle, as returned from :meth:`scatter`.

        patch
            Some kind of Python object. Must be serializable
            and must match the :code:`obj.patch()` method.
        '''
        raise NotImplementedError()

    @contextmanager
    def scatter(self, obj):
        '''
        Scatter :code:`obj` throughout the cluster

        Parameters
        ----------

        obj
            Some kind of Python object. Must be serializable.

        Returns
        -------
        handle
            Handle for the scattered :code:`obj`
        '''
        raise NotImplementedError()

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: "TaskCommHandler",
    ):
        """
        Run the tasks with the given parameters.

        Raises
        ------
        JobCancelledError
            Either the job was cancelled using :meth:`AsyncJobExecutor.cancel`,
            or the underlying data source was interrupted.

        Parameters
        ----------
        tasks
            The tasks to be run
        params_handle : [type]
            A handle for the task parameters, as returned from :meth:`scatter`
        cancel_id
            An identifier which can be used for cancelling all tasks together. The
            same identifier should be passed to :meth:`AsyncJobExecutor.cancel`
        """
        raise NotImplementedError()

    def run_each_partition(self, partitions, fn, all_nodes=False):
        """
        Run `fn` for all partitions. Yields results in order of completion.

        Parameters
        ----------

        partitions : List[Partition]
            List of relevant partitions.

        fn : callable
            Function to call, will get the partition as first and only argument.

        all_nodes : bool
            If all_nodes is True, run the function on all nodes that have this partition,
            otherwise run on any node that has the partition. If a partition has no location,
            the function will not be run for that partition if `all_nodes` is True, otherwise
            it will be run on any node.
        """
        raise NotImplementedError()

    def map(self, fn: Callable[[V], T], iterable: Iterable[V]) -> Iterable[T]:
        """
        Run a callable :code:`fn` for each element in :code:`iterable`, on arbitrary worker nodes.

        Parameters
        ----------

        fn : callable
            Function to call. Should accept exactly one parameter.

        iterable : Iterable
            Which elements to call the function on.
        """
        raise NotImplementedError()

    def run_each_host(self, fn, *args, **kwargs):
        """
        Run a callable :code:`fn` once on each host, gathering all results into
        a dict host -> result

        Parameters
        ----------

        fn : callable
            Function to call

        *args
            Arguments for fn

        **kwargs
            Keyword arguments for fn
        """
        raise NotImplementedError()

    def run_each_worker(self, fn, *args, **kwargs):
        """
        Run :code:`fn` on each worker process, and pass :code:`*args`,
        :code:`**kwargs` to it.

        Useful, for example, if you need to prepare the environment of
        each Python interpreter, warm up per-process caches etc.

        Parameters
        ----------
        fn : callable
            Function to call
        \\*args
            Arguments for fn
        \\*\\*kwargs
            Keyword arguments for fn

        Returns
        -------
        dict
            Return values keyed by worker name (executor-specific)
        """
        raise NotImplementedError()

    def close(self):
        """
        cleanup resources used by this executor, if any
        """

    def get_available_workers(self) -> WorkerSet:
        """
        Returns a WorkerSet that contains the available workers

        Each worker should correspond to a "worker process", so if the executor
        is using multiple processes or threads, each process/thread should be
        included in this list.
        """
        raise NotImplementedError()

    def get_resource_details(self) -> list[dict[str, Any]]:
        """
        Returns a list of dicts with cluster details

        key of the dictionary:
            host: ip address or hostname where the worker is running
        """
        raise NotImplementedError()

    def ensure_sync(self):
        """
        Returns a synchronous executor. In case of a
        :class:`~libertem.common.executor.JobExecutor` we just return
        :code:`self`; in case of
        :class:`~libertem.common.executor.AsyncJobExecutor` below more work is
        needed!
        """
        return self

    def ensure_async(self, pool=None):
        """
        Returns an asynchronous executor; by default just wrap into
        :class:`~libertem.executor.base.AsyncAdapter`.
        """
        raise NotImplementedError()

    def modify_buffer_type(self, buf):
        """
        Allow the executor to modify result buffers if necessary

        Currently only called for buffers on the main node
        """
        return buf

    def get_udf_runner(self) -> type['UDFRunner']:
        raise NotImplementedError


class AsyncJobExecutor:
    '''
    Async version of :class:`JobExecutor`.
    '''
    async def run_tasks(self, tasks, params_handle, cancel_id):
        """
        Run a number of Tasks, yielding (result, task) tuples
        """
        raise NotImplementedError()

    async def run_function(self, fn: Callable[..., T], *args, **kwargs) -> T:
        """
        Run a callable :code:`fn` on any worker
        """
        raise NotImplementedError()

    async def run_each_partition(self, partitions, fn, all_nodes=False):
        raise NotImplementedError()

    async def map(self, fn, iterable):
        """
        Run a callable :code:`fn` for each item in :code:`iterable`, on arbitrary worker nodes

        Parameters
        ----------

        fn : callable
            Function to call. Should accept exactly one parameter.

        iterable : Iterable
            Which elements to call the function on.
        """
        raise NotImplementedError()

    async def run_each_host(self, fn, *args, **kwargs):
        raise NotImplementedError()

    async def run_each_worker(self, fn, *args, **kwargs):
        """
        Run :code:`fn` on each worker process, and pass :code:`*args`, :code:`**kwargs` to it.

        Useful, for example, if you need to prepare the environment of
        each Python interpreter, warm up per-process caches etc.

        Parameters
        ----------
        fn : callable
            Function to call
        \\*args
            Arguments for fn

        \\*\\*kwargs
            Keyword arguments for fn
        """
        raise NotImplementedError()

    async def close(self):
        """
        Cleanup resources used by this executor, if any.
        """

    async def cancel(self, cancel_id):
        """
        cancel execution identified by `cancel_id`
        """
        pass

    async def get_available_workers(self):
        raise NotImplementedError()

    async def get_resource_details(self):
        raise NotImplementedError()

    def ensure_sync(self):
        raise NotImplementedError()

    def ensure_async(self, pool=None):
        """
        Returns an asynchronous executor; by default just return `self`.
        """
        return self

    def get_udf_runner(self) -> type['UDFRunner']:
        raise NotImplementedError()


class WorkerQueueEmpty(Exception):
    """
    A non-blocking get was called on an empty queue, or a blocking `get` with a
    non-zero timeout timed out.
    """
    pass


class WorkerQueue:
    '''
    Interface for queues to send input data to workers.
    '''
    @contextmanager
    def get(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
    ) -> Generator[tuple[Any, memoryview], None, None]:
        raise NotImplementedError()

    def put(self, header: Any, payload: Optional[memoryview] = None):
        '''
        Put header and payload into the queue.
        '''
        raise NotImplementedError()

    @contextmanager
    def put_nocopy(self, header: Any, size: int) -> Generator[memoryview, None, None]:
        """
        Put data into the queue, without an additional copy. This will yield a
        writable :class:`python:memoryview`, instead of requiring that the data
        to be sent is already available as a :class:`python:memoryview` and
        copying into its final destination, as is the case in :meth:`put`.

        This can be useful, for example, if you need to perform any kind of
        operation that writes into a buffer. For example:

        >>> q = SomeWorkerQueueImpl()  # doctest: +SKIP
        >>> with q.put_nocopy(1024) as send_buf:  # doctest: +SKIP
        ...     # NOTE: in reality, need to handle n_bytes < 1024
        ...     n_bytes = some_socket.recvinto(send_buf)

        If the data you receive is already in a shared memory segment, you can
        send the handle as part of the :code:`header` and use the normal
        :meth:`put` method instead.

        Note: for releasing memory, it might be necessary to implement another
        queue that synchronizes the release process.
        """
        raise NotImplementedError()

    def close(self, drain: bool = True, force: bool = False):
        """
        Parameters
        ----------

        drain
            If needed by the underlying queue, remove any items
            from the queue before closing

        force
            Don't wait for data to be flushed, forcefully close the queue
        """
        raise NotImplementedError()

    def size(self) -> int:
        """
        Approximate number of items currently in the queue

        Can raise `NotImplementedError` depending on operating system
        or underlying implementation.
        """
        raise NotImplementedError()


class SimpleWorkerQueue(WorkerQueue):
    """
    A :class:`WorkerQueue` that uses a threading :class:`python:queue.Queue`
    under the hood.
    """
    def __init__(self) -> None:
        self.q: queue.Queue = queue.Queue()

    def put(self, header, payload: Optional[memoryview] = None):
        self.q.put((header, payload))

    @contextmanager
    def put_nocopy(self, header: Any, size: int) -> Generator[memoryview, None, None]:
        payload = np.zeros(size, dtype=np.uint8)
        yield memoryview(payload)
        self.q.put((header, payload))

    @contextmanager
    def get(self, block: bool = True, timeout: Optional[float] = None):
        try:
            res = self.q.get(block=block, timeout=timeout)
            yield res
        except queue.Empty:
            raise WorkerQueueEmpty()

    def close(self, drain: bool = True, force: bool = False):
        pass

    def size(self) -> int:
        return self.q.qsize()


class SimpleMPWorkerQueue(WorkerQueue):
    """
    A :class:`WorkerQueue` that uses a :class:`python:multiprocessing.Queue` under the hood.
    """
    def __init__(self) -> None:
        self._mp_ctx = mp.get_context("spawn")
        self.q: mp.Queue = self._mp_ctx.Queue()
        self._closed = False

    def put(self, header, payload: Optional[memoryview] = None):
        with tracer.start_as_current_span("SimpleMPWorkerQueue.put") as span:
            header_serialized = cloudpickle.dumps(header)
            payload_serialized = cloudpickle.dumps(payload)
            span.set_attributes({
                'libertem.pickle.header_size': len(header_serialized),
                'libertem.pickle.payload_size': len(payload_serialized),
            })
            self.q.put((header_serialized, payload_serialized))

    @contextmanager
    def put_nocopy(self, header: Any, size: int) -> Generator[memoryview, None, None]:
        payload = np.zeros(size, dtype=np.uint8)
        yield memoryview(payload)
        header_serialized = cloudpickle.dumps(header)
        payload_serialized = cloudpickle.dumps(payload)
        self.q.put((header_serialized, payload_serialized))

    @contextmanager
    def get(self, block: bool = True, timeout: Optional[float] = None):
        try:
            res = self.q.get(block=block, timeout=timeout)
            yield (cloudpickle.loads(res[0]), cloudpickle.loads(res[1]))
        except queue.Empty:
            raise WorkerQueueEmpty()

    def close(self, drain: bool = True, force: bool = False):
        if not self._closed:
            if drain:
                while True:
                    try:
                        self.q.get_nowait()
                    except queue.Empty:
                        break
            self.q.close()
            if force:
                self.q.cancel_join_thread()
            else:
                self.q.join_thread()
            self._closed = True

    def size(self) -> int:
        return self.q.qsize()


class WorkerContext:
    """
    A :class:`WorkerContext` is used to manage streaming communication between
    the main process and the workers.
    """
    def get_worker_queue(self) -> WorkerQueue:
        raise NotImplementedError()

    def signal(self, ident: str, topic: str, msg_dict: dict[str, Any]):
        raise NotImplementedError()


class TaskCommHandler:
    """
    This is the interface that is implemented by the acquisition object
    to allow streaming communication with workers.
    """
    def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
        """
        Handle the streaming connunication for the given :code:`task`
        using the provided :code:`queue`. This function should
        block until the communication for the given task has finished.
        It may be run in a background thread on the main node,
        or synchronously, depending on the :class:`JobExecutor`.

        May raise :class:`JobCancelledError` to signal that the acquisition has
        been cancelled for some reason.

        Parameters
        ----------
        task : TaskProtocol

        queue : WorkerQueue
            A queue used to communicate with the worker process
            using an acquisition-specific protocol. The `Partition`
            in the worker has access to this queue, too, and can
            communicate using a data-source specific protocol.

            The `TaskCommHandler` and the `Partition` are tightly
            coupled by this protocol, and the queue needs to be "clean"
            after the data for a given task has been exchanged.
        """
        ...

    def start(self):
        """
        A lifecycle method that is called before any task has been
        run.
        """
        ...

    def done(self):
        """
        A lifecycle method that is called after all tasks have benn
        run.
        """
        ...

    @property
    def subscriptions(self) -> dict[str, list[Callable]]:
        # Instantiate on first get to avoid creating __init__
        try:
            return self._subscriptions
        except AttributeError:
            self._subscriptions = {}
            return self._subscriptions

    def subscribe(self, topic: str, callback: Callable[[str, dict], None]):
        """
        Register a callback to run in response to messages
        matching the topic string identifier

        The callback should accept the arguments
            (topic, message_dict)
        message_dict will contain an 'ident' key
        with the identiy of the message sender
        """
        try:
            self.subscriptions[topic].append(callback)
        except KeyError:
            self.subscriptions[topic] = [callback]

    @contextmanager
    def monitor(self, queue: WorkerQueue):
        """
        Monitor queue in a background thread and run the
        callbacks in subscriptions in response to messages
        """
        # Avoid circular import
        from libertem.common.progress import CommsDispatcher
        with CommsDispatcher(queue, self.subscriptions):
            yield


class NoopCommHandler(TaskCommHandler):
    """
    A :class:`TaskCommHandler` that doesn't perform any action, and doesn't
    stream any data.
    """
    def handle_task(self, task: TaskProtocol, queue: WorkerQueue):
        pass

    def start(self):
        pass

    def done(self):
        pass



import io
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import pytest

from libertem.web.notebook_generator.notebook_generator import notebook_generator
from libertem.io.dataset.base import BufferedBackend


@pytest.mark.slow
def test_custom_io_backend(buffered_raw, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = buffered_raw._path

    dataset_params = {
        "type": "RAW",
        "params": {
            "path": path,
            "dtype": "float32",
            "nav_shape": [16, 16],
            "sig_shape": [128, 128],
            "io_backend": BufferedBackend(),
        }
    }

    analysis = [{
        "analysisType": 'APPLY_DISK_MASK',
        "parameters": {
            'shape': 'disk',
            'cx': 8,
            'cy': 8,
            'r': 5,
        }
    }]

    notebook = notebook_generator(conn, dataset_params, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    # this throws an exception if the `io_backend` parameter is not generated
    # properly in the notebook:
    ep.preprocess(nb, {"metadata": {"path": datadir}})



import pytest
import numpy as np

from libertem.udf.holography import HoloReconstructUDF
from libertem.io.dataset.memory import MemoryDataSet
from libertem.utils.generate import hologram_frame
from libertem.utils.devices import detect
from libertem.common.backend import set_use_cpu, set_use_cuda


@pytest.mark.parametrize(
    # CuPy support deactivated due to https://github.com/LiberTEM/LiberTEM/issues/815
    # 'backend', ['numpy', 'cupy']
    'backend', ['numpy']
)
def test_holo_reconstruction(lt_ctx, backend):
    if backend == 'cupy':
        d = detect()
        cudas = detect()['cudas']
        if not d['cudas'] or not d['has_cupy']:
            pytest.skip("No CUDA device or no CuPy, skipping CuPy test")
    # Prepare image parameters and mesh
    nx, ny = (5, 7)
    sx, sy = (64, 64)
    slice_crop = (slice(None),
                  slice(None),
                  slice(sx // 4, sx // 4 * 3),
                  slice(sy // 4, sy // 4 * 3))

    lnx = np.arange(nx)
    lny = np.arange(ny)
    lsx = np.arange(sx)
    lsy = np.arange(sy)

    mnx, mny, msx, msy = np.meshgrid(lnx, lny, lsx, lsy)

    # Prepare phase image
    phase_ref = np.pi * msx * (mnx.max() - mnx) * mny / sx**2 \
        + np.pi * msy * mnx * (mny.max() - mny) / sy**2

    # Generate holograms
    holo = np.zeros_like(phase_ref)
    ref = np.zeros_like(phase_ref)

    for i in range(nx):
        for j in range(ny):
            holo[j, i, :, :] = hologram_frame(np.ones((sx, sy)), phase_ref[j, i, :, :])
            ref[j, i, :, :] = hologram_frame(np.ones((sx, sy)), np.zeros((sx, sy)))

    # Prepare LT datasets and do reconstruction
    dataset_holo = MemoryDataSet(data=holo, tileshape=(ny, sx, sy),
                                 num_partitions=2, sig_dims=2)

    dataset_ref = MemoryDataSet(data=ref, tileshape=(ny, sx, sy),
                                num_partitions=1, sig_dims=2)

    sb_position = [11, 6]
    sb_size = 6.26498204

    holo_job = HoloReconstructUDF(out_shape=(sx, sy),
                                  sb_position=sb_position,
                                  sb_size=sb_size)
    try:
        if backend == 'cupy':
            set_use_cuda(cudas[0])
        w_holo = lt_ctx.run_udf(dataset=dataset_holo, udf=holo_job)['wave'].data
        w_ref = lt_ctx.run_udf(dataset=dataset_ref, udf=holo_job)['wave'].data
    finally:
        set_use_cpu(0)

    w = w_holo / w_ref

    phase = np.angle(w)

    assert np.allclose(phase_ref[slice_crop], phase[slice_crop], rtol=0.12)


def test_sb_pos_invalid(lt_ctx):
    with pytest.raises(ValueError) as e:
        HoloReconstructUDF(
            out_shape=(128, 128),
            sb_position=(0, 1, 2),
            sb_size=(0, 1, 2)
        )

    assert e.match(r"invalid sb_position \(0, 1, 2\), must be tuple of length 2")



import pytest

from libertem.common.async_utils import (
    sync_to_async, MyStopIteration, run_agen_get_last, run_gen_get_last,
)


def _f():
    print("stuff while")
    next(iter([]))


@pytest.mark.asyncio
async def test_sync_to_async_stopiteration():
    print("stuff before")
    with pytest.raises(MyStopIteration):
        await sync_to_async(_f)


@pytest.mark.asyncio
async def test_run_agen_get_last_empty():
    async def _empty_agen():
        for i in []:
            yield i

    with pytest.raises(RuntimeError):
        await run_agen_get_last(_empty_agen())


def test_run_gen_get_last_empty():
    def _empty_gen():
        yield from []

    with pytest.raises(RuntimeError):
        run_gen_get_last(_empty_gen())



import sys
import queue
import asyncio
import threading
from typing import Callable, Optional, TypeVar
from collections.abc import AsyncGenerator, Generator
from concurrent.futures import ThreadPoolExecutor

from opentelemetry import trace

from libertem.common.tracing import attach_to_parent

tracer = trace.get_tracer(__name__)
T = TypeVar('T')


class MyStopIteration(Exception):
    """
    TypeError: StopIteration interacts badly with generators
    and cannot be raised into a Future
    """
    pass


def _wrap_f(f, args, kwargs):
    def _wrapped():
        try:
            return f(*args, **kwargs)
        except StopIteration as e:
            raise MyStopIteration() from e
    return _wrapped


async def sync_to_async(
    fn: Callable[..., T], pool: Optional[ThreadPoolExecutor] = None, *args, **kwargs
) -> T:
    """
    Run blocking function with `*args`, `**kwargs` in a thread pool.

    Raises `MyStopIteration` in case `fn` raises a `StopIteration`.

    Parameters
    ----------
    fn : callable
        The blocking function to run in a background thread

    pool : ThreadPoolExecutor or None
        In which thread pool should the function be run? If `None`, we create a new one

    *args, **kwargs
        Passed on to `fn`
    """
    loop = asyncio.get_event_loop()
    # As `fn` may throw a `StopIteration` for whatever reason, which doesn't
    # work inside of futures, we have to wrap it in a custom exception type.
    # Without this, it's possible that things hang instead of propagating
    # errors properly:
    fn = _wrap_f(fn, args, kwargs)
    return await loop.run_in_executor(pool, fn)


async def async_generator(
    gen: Generator[T, None, None], pool: Optional[ThreadPoolExecutor] = None
) -> AsyncGenerator[T, None]:
    def inner_next(gen):
        try:
            return next(gen)
        except StopIteration:
            raise MyStopIteration()

    loop = asyncio.get_event_loop()
    while True:
        try:
            item = await loop.run_in_executor(pool, inner_next, gen)
            yield item
        except MyStopIteration:
            break


async def run_agen_get_last(gen: AsyncGenerator[T, None]) -> T:
    result = None
    try:
        while True:
            result = await gen.__anext__()
    except StopAsyncIteration:
        pass
    finally:
        await gen.aclose()
    if result is None:
        raise RuntimeError("run_agen_get_last called with empty generator")
    return result


def run_gen_get_last(gen: Generator[T, None, None]) -> T:
    result = None
    try:
        while True:
            result = gen.__next__()
    except StopIteration:
        pass
    finally:
        gen.close()
    if result is None:
        raise RuntimeError("run_gen_get_last called with empty generator")
    return result


class AsyncGenToQueueThread(threading.Thread):
    """
    Wrap an async generator and execute it in a thread, putting the generated
    items into a queue.

    Parameters
    ----------
    agen : iterable
        The async generator to wrap

    q : queue.Queue
        The result queue where the generated items will be put. The calling
        thread should consume the items in this queue.

    args, kwargs
        will be passed to :code:`Thread.__init__`
    """
    def __init__(self, agen, q, span_context, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._agen = agen
        self._q = q
        self._should_stop = threading.Event()
        self._span_context = span_context
        self.ex = None

    def run(self):
        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            adjust_event_loop_policy()

            async def _wrap_gen(agen):
                async for item in agen:
                    self._q.put(item)
                    if self._should_stop.is_set():
                        break
            loop.run_until_complete(_wrap_gen(self._agen))
        except Exception as e:
            self.ex = e
        finally:
            self._q.put(MyStopIteration)
        return

    def get_exception(self):
        return self.ex

    def stop(self):
        self._should_stop.set()


def async_to_sync_generator(agen, pool=None):
    span = trace.get_current_span()
    span_context = span.get_span_context()
    q = queue.Queue()
    t = AsyncGenToQueueThread(agen, q, span_context)
    t.start()
    try:
        while True:
            item = q.get()
            if item is MyStopIteration:
                # propagate any uncaught exception in the wrapped generator to the calling thread:
                ex = t.get_exception()
                if ex is not None:
                    raise ex
                break
            yield item

        if q.qsize() > 0:
            res = q.get()
            if isinstance(res, Exception):
                raise res  # TODO: re-raise in a better way?
    finally:
        t.join()


class SyncGenToQueueThread(threading.Thread):
    """
    Wrap a generator and execute it in a thread, putting the generated
    items into a queue.

    Parameters
    ----------
    gen : iterable
        The generator to wrap

    q : queue.Queue
        The result queue where the generated items will be put. The calling
        thread should consume the items in this queue.

    args, kwargs
        will be passed to :code:`Thread.__init__`
    """
    def __init__(self, gen, q, span_context, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._gen = gen
        self._q = q
        self._should_stop = threading.Event()
        self._span_context = span_context
        self.ex = None

    def run(self):
        with attach_to_parent(self._span_context):
            try:
                # looping over `self._gen` can take some time for each item,
                # this is where the "real" work happens:
                for item in self._gen:
                    self._q.put(item)
                    if self._should_stop.is_set():
                        break
            except Exception as e:
                self.ex = e
            finally:
                self._q.put(MyStopIteration)
            return

    def get_exception(self):
        return self.ex

    def stop(self):
        self._should_stop.set()


async def async_generator_eager(gen, pool=None):
    """
    Convert the synchronous generator `gen` to an async generator. Eagerly run
    `gen` in a thread and provide the result in a queue. This means that `gen`
    can run ahead of the asynchronous generator that is returned.

    Parameters:
    -----------

    gen : iterable
        The generator to run

    pool: ThreadPoolExecutor
        The thread pool to run the generator in, can be None to create an
        ad-hoc thread
    """
    span = trace.get_current_span()
    span_context = span.get_span_context()
    q = queue.Queue()
    t = SyncGenToQueueThread(gen, q, span_context)

    loop = asyncio.get_event_loop()

    t.start()
    try:
        while True:
            # get a single item from the result queue:
            item = await loop.run_in_executor(pool, q.get)

            # MyStopIteration is a canary value to signal that the inner
            # generator has finished running
            if item is MyStopIteration:
                # propagate any uncaught exception in the wrapped generator to the calling thread:
                ex = t.get_exception()
                if ex is not None:
                    raise ex
                break
            yield item
    finally:
        # in case our thread raises an exception, we may need to stop the `SyncGenToQueueThread`:
        t.stop()
        t.join()


def adjust_event_loop_policy():
    """
    Set an appropriate event loop policy on Windows. The new one from Python 3.8 doesn't
    work for us by default, so call this as early as possible!
    """
    # stolen from ipykernel:
    if sys.platform.startswith("win"):
        try:
            from asyncio import (
                WindowsProactorEventLoopPolicy,
                WindowsSelectorEventLoopPolicy,
            )
        except ImportError:
            # not affected
            pass
        else:
            # FIXME this might fail if the event loop policy has been overridden by something
            # incompatible that is not a WindowsProactorEventLoopPolicy. Dask.distributed creates
            # a custom event loop policy, for example. Fortunately that one should be compatible.
            # Better would be "duck typing", i.e. to check if the required add_reader() method
            # of the loop throws a NotImplementedError.
            if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                # fallback to the pre-3.8 default of Selector
                asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())



import os
import numpy as np
import io
import nbformat
from temp_utils import _get_hdf5_params
from libertem.udf.sum import SumUDF
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from libertem import masks
from nbconvert.preprocessors import ExecutePreprocessor
import pytest


@pytest.mark.slow
def test_sum_default(hdf5_ds_2, tmpdir_factory, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)
    analysis = [{
            "analysisType": "SUM_FRAMES",
            "parameters": {
                    "roi": {}
                    }
            }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'sum_result.npy')
    result = np.load(data_path)
    with hdf5_ds_2.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
    expected = data.sum(axis=(0, 1))
    assert np.allclose(expected, result)


@pytest.mark.slow
def test_sum_roi(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)
    roi_params = {
        "shape": "disk",
        "cx": 8,
        "cy": 8,
        "r": 6
    }
    analysis = [{
                "analysisType": "SUM_FRAMES",
                "parameters": {
                            "roi": roi_params
                            }
                }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'sum_result.npy')
    results = np.load(data_path)
    nx, ny = hdf5_ds_2.shape.nav
    roi = masks.circular(
                    centerX=roi_params['cx'],
                    centerY=roi_params['cy'],
                    imageSizeX=nx,
                    imageSizeY=ny,
                    radius=roi_params['r'],
            )
    udf = SumUDF()
    expected = lt_ctx.run_udf(hdf5_ds_2, udf, roi)

    assert np.allclose(
        results,
        expected['intensity'].raw_data,
    )



import pytest
import numpy as np
from scipy.ndimage import gaussian_filter
import libertem.utils as ut
from libertem.utils.generate import hologram_frame


def test_polar():
    data = np.array([
        [(0, 1), (0, 1)],
        [(1, 0), (2, 0)],
        [(-2, 0), (0, 1)],
    ])
    expected = np.array([
        [(1, 0), (1, 0)],
        [(1, np.pi/2), (2, np.pi/2)],
        [(2, -np.pi/2), (1, 0)],
    ])

    result = ut.make_polar(data)
    assert data.shape == expected.shape
    assert result.shape == expected.shape
    assert np.allclose(expected, result)


def test_conversion(points):
    assert np.allclose(points, ut.make_cartesian(ut.make_polar(points)))


def test_rotation():
    y, x = -1, 2
    r_y, r_x = ut.rotate_deg(y, x, 90)
    # Clock-wise with y pointing down
    assert np.allclose(r_y, 2)
    assert np.allclose(r_x, 1)


def test_rotate_multi():
    y, x = np.random.random(2) - 0.5
    angle = 30
    cos_angle = np.cos(np.pi*angle/180)
    sin_angle = np.sin(np.pi*angle/180)
    assert 360 % angle == 0
    r_y = y
    r_x = x
    for i in range(360//angle):
        r_y2, r_x2 = ut.rotate_deg(
            y=r_y,
            x=r_x,
            degrees=angle
        )
        r_y, r_x = ut.rotate_precalc(
            y=r_y,
            x=r_x,
            cos_angle=cos_angle,
            sin_angle=sin_angle
        )
        assert np.allclose(r_y, r_y2)
        assert np.allclose(r_x, r_x2)
    print(r_y, r_x)
    assert np.allclose(r_y, y)
    assert np.allclose(r_x, x)


def test_polar_rotate():
    y, x, radians = np.random.random(3)
    ((r, phi),) = ut.make_polar(np.array([(y, x)]))
    ((r_y, r_x),) = ut.make_cartesian(np.array([(r, phi + radians)]))
    r_y2, r_x2 = ut.rotate_rad(y, x, radians)
    assert np.allclose(r_y, r_y2)
    assert np.allclose(r_x, r_x2)


@pytest.mark.parametrize('counts, sampling, visibility, f_angle, gauss, poisson, rtol1, rtol2',
                         [(None, None, None, None, None, None, 1e-5, 2e-2),
                          (500., 6.2, 0.5, 66., 0.7, 1e-4, 6e-2, .3)])
def test_hologram_frame(counts, sampling, visibility, f_angle, gauss, poisson, rtol1, rtol2):
    sx, sy = (32, 64)
    x, y = np.meshgrid(np.arange(sx), np.arange(sy))

    kwargs = {}
    if counts:
        kwargs['counts'] = counts
    else:
        counts = 1000.
    if sampling:
        kwargs['sampling'] = sampling
    else:
        sampling = 5.
    if visibility:
        kwargs['visibility'] = visibility
    else:
        visibility = 1.
    if f_angle:
        kwargs['f_angle'] = f_angle
    else:
        f_angle = 30.
    if gauss:
        kwargs['gaussian_noise'] = gauss
    if poisson:
        kwargs['poisson_noise'] = poisson
    amp = 1 - np.random.randn(sy, sx) * 0.01
    phase = np.random.random() * x / sx + np.random.random() * y / sy
    holo_test = hologram_frame(amp, phase, **kwargs)

    f_angle = f_angle / 180. * np.pi

    holo = counts / 2 * (1. + amp ** 2 + 2. * amp * visibility
                         * np.cos(2. * np.pi * y / sampling * np.cos(f_angle)
                                  + 2. * np.pi * x / sampling * np.sin(f_angle)
                                  - phase))

    if poisson:
        noise_scale = poisson * counts
        holo = noise_scale * np.random.poisson(holo / noise_scale)

    if gauss:
        holo = gaussian_filter(holo, gauss)
    assert np.allclose(holo, holo_test, rtol=rtol1)

    # test derived parameters:
    # test if mean value equals to counts
    assert np.isclose(holo_test.mean(), counts, rtol=5e-3)

    # test if calculated contrast is equals to teh input
    def contrast(a):
        return (a.max(1).mean() - a.min(1).mean()) \
                / (a.min(1).mean() + a.max(1).mean())
    assert np.isclose(contrast(holo_test), visibility, rtol=rtol2)

    # test if fringe spacing equals to the input
    holo_fft = np.abs(np.fft.rfft2(holo_test[:sx, :sx]))
    holo_fft[:1, :1] = 0.
    holo_max = np.unravel_index(holo_fft.argmax(), holo_fft.shape)
    holo_max = np.hypot(holo_max[0], holo_max[1])
    sampling_test = sx / holo_max
    error_sampling = sampling_test * (1. / holo_max)
    assert np.isclose(sampling_test, sampling, atol=error_sampling)


def test_holo_frame_asserts():
    # test asserts:
    with pytest.raises(ValueError):
        hologram_frame(np.ones((7, 5)), np.zeros((5, 7)))
    with pytest.raises(ValueError):
        hologram_frame(np.ones((5, 7)), np.zeros((5, 7)), gaussian_noise='a lot')
    with pytest.raises(ValueError):
        hologram_frame(np.ones((5, 7)), np.zeros((5, 7)), poisson_noise='a bit')



import os
import re
import csv
from glob import glob, escape
import typing
import logging
import warnings
import configparser

import numba
from numba.typed import List
import scipy.io as sio
import numpy as np

from libertem.common.math import prod
from libertem.io.corrections import CorrectionSet
from libertem.common import Shape, Slice
from libertem.common.math import flat_nonzero
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, DataSetException, DataSetMeta,
    FileSet, BasePartition, File, Decoder,
    TilingScheme, make_get_read_ranges, IOBackend,
)

log = logging.getLogger(__name__)
READOUT_MODE_PAT = re.compile(
    r'^"bin:\s*(?P<bin>\d+),\s*windowing:\s*(?P<win_i>\d+)\s*x\s*(?P<win_j>\d+)\s*"$'
)

# file header: 1024 bytes, at the beginning of each file
file_header_dtype = [
    ('header_size', '<u2'),        # fixed value: 1024
    ('frame_header_size', '<u2'),  # fixed value: 64
    ('padding_1', (bytes, 3)),
    ('version', '<u1'),            # fixed value: 6
    ('comment_1', (bytes, 80)),
    ('width', '<u2'),              # -> columns
    ('height', '<u2'),             # -> rows
    ('comment_2', (bytes, 928)),
    ('num_frames', '<u4'),
]

# frame header: 64 bytes, before each frame
frame_header_dtype = [
    ('padding_1', (bytes, 4)),
    ('timestamp_s', '<u4'),
    ('timestamp_us', '<u4'),
    ('frame_number', '<u4'),
    ('padding_2', (bytes, 12)),
    ('comment', (bytes, 36)),
]


class FRMS6DatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/FRMS6DatasetParams.schema.json",
        "title": "FRMS6DatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "FRMS6"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
        },
        "required": ["type", "path"]
        }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class GainMapCSVDialect(csv.excel):
    delimiter = ';'


def _unbin(tile_data, factor):
    """
    tile_data should have shape (num_frames, y, x)
    """
    s = tile_data.shape

    # insert a binning dimension:
    tile_data = tile_data.reshape((s[0], s[1], 1, s[2]))
    unbinned = tile_data.repeat(factor, axis=2)
    # FIXME: should we scale the data by the binning factor?
    return unbinned.reshape((s[0], factor * s[1], s[2]))


def _get_base_filename(_path):
    path, ext = os.path.splitext(_path)
    if ext == ".hdr":
        base = path
    elif ext == ".frms6":
        base = re.sub(r'_[0-9]+$', '', path)
    else:
        raise DataSetException("unknown extension: %s" % ext)
    return base


def _read_dataset_hdr(fname):
    if not os.path.exists(fname):
        raise DataSetException(
            "Could not find .hdr file {}".format(
                fname,
            )
        )

    config = configparser.ConfigParser()
    config.read(fname)
    parsed = {}
    sections = config.sections()
    if 'measurementInfo' not in sections:
        raise DataSetException(
            "measurementInfo missing from .hdr file {}, have: {}".format(
                fname,
                repr(sections),
            )
        )
    msm_info = config['measurementInfo']
    int_fields = {'darkframes', 'dwelltimemicroseconds', 'gain', 'signalframes'}
    for key in msm_info:
        value = msm_info[key]
        if key in int_fields:
            parsed[key] = int(value)
        else:
            parsed[key] = value
    # FIXME: are the dimensions the right way aroud? is there a sample file with a non-square
    # scan region?
    parsed['stemimagesize'] = tuple(int(p) for p in parsed['stemimagesize'].split('x'))
    match = READOUT_MODE_PAT.match(parsed['readoutmode'])
    if match is None:
        raise DataSetException("could not parse readout mode")
    readout_mode = match.groupdict()
    parsed['readoutmode'] = {k: int(v) for k, v in readout_mode.items()}
    return parsed


def _read_file_header(path):
    header_raw = np.fromfile(path, dtype=file_header_dtype, count=1)
    header = {}
    for field, dtype in file_header_dtype:
        if type(dtype) is not str:
            continue
        header[field] = header_raw[field][0]
        # to prevent overflows in following computations:
        if np.dtype(dtype).kind == "u":
            header[field] = int(header[field])
    header['filesize'] = os.stat(path).st_size
    header['path'] = path
    return header


def _pattern(path):
    path, ext = os.path.splitext(path)
    if ext == ".hdr":
        pattern = "%s_*.frms6" % escape(path)
    elif ext == ".frms6":
        pattern = "%s*.frms6" % (
            escape(re.sub(r'[0-9]+$', '', path))
        )
    else:
        raise DataSetException("unknown extension: %s" % ext)
    return pattern


def get_filenames(path, disable_glob=False):
    if disable_glob:
        return [path]
    else:
        return list(sorted(glob(_pattern(path))))


def _get_sig_shape(path, bin_factor):
    filenames = get_filenames(path)
    header = _read_file_header(filenames[0])
    sig_shape = 2 * header['height'], header['width'] // 2
    if bin_factor > 1:
        sig_shape = (sig_shape[0] * bin_factor, sig_shape[1])
    return sig_shape


def _header_valid(header):
    if not (header['header_size'] == 1024
            and header['frame_header_size'] == 64
            and header['version'] == 6):
        return False
    # TODO: file size sanity check?
    return True


def _num_frames(header):
    if header['num_frames'] != 0:
        return header['num_frames']
    # older FRMS6 files don't contain the number of frames in the header,
    # so calculate from filesize:
    w, h = header['width'], header['height']
    bytes_per_frame = w * h * 2
    assert header['filesize'] is not None
    num = (header['filesize'] - 1024)
    denum = (bytes_per_frame + 64)
    res = num // denum
    if num % denum != 0:
        raise DataSetException("could not determine number of frames")
    return res


@numba.njit(inline='always')
def _map_y(y, xs, binning, num_rows):
    """
    Parameters
    ----------

    binning : int
        Binning factor (1, 2, 4)

    num_rows : int
        Total number of rows (should be 264, if not windowing)
    """
    half = num_rows // 2 // binning
    if y < half:
        return (y, 0)
    else:
        return (((num_rows // binning) - y - 1), xs)


@numba.njit(inline='always')
def _frms6_read_ranges_tile_block(
    slices_arr, fileset_arr, slice_sig_sizes, sig_origins,
    inner_indices_start, inner_indices_stop, frame_indices, sig_size,
    px_to_bytes, bpp, frame_header_bytes, frame_footer_bytes, file_idxs,
    slice_offset, extra, sig_shape,
):
    """
    Read ranges for frms6: create read ranges line-by-line
    """
    result = List()

    # positions in the signal dimensions:
    for slice_idx in range(slices_arr.shape[0]):
        # (offset, size) arrays defining what data to read (in pixels)
        slice_origin = slices_arr[slice_idx][0]
        slice_shape = slices_arr[slice_idx][1]

        read_ranges = List()

        x_shape = slice_shape[1]
        stride = 2 * x_shape

        binning = extra[0]
        num_rows = sig_shape[0]

        # inner "depth" loop along the (flat) navigation axis of a tile:
        for i, inner_frame_idx in enumerate(range(inner_indices_start, inner_indices_stop)):
            inner_frame = frame_indices[inner_frame_idx]
            file_idx = file_idxs[i]
            f = fileset_arr[file_idx]
            frame_in_file_idx = inner_frame - f[0]
            file_header_bytes = f[3]

            # we are reading a part of a single frame, so we first need to find
            # the offset caused by headers:
            header_offset = file_header_bytes + frame_header_bytes * (frame_in_file_idx + 1)

            # now let's figure in the current frame index:
            # (go down into the file by full frames; `sig_size`)
            offset = header_offset + frame_in_file_idx * (sig_size // binning) * bpp // 8

            y_start = slice_origin[0] // binning
            y_stop = (slice_origin[0] + slice_shape[0]) // binning

            for y in range(y_start, y_stop):
                mapped_y, x_offset = _map_y(y, x_shape, binning, num_rows)
                start = offset + (stride * mapped_y + x_offset) * bpp // 8
                read_ranges.append(
                    (
                        file_idx,
                        start, start + x_shape * bpp // 8
                    )
                )

        # the indices are compressed to the selected frames
        compressed_slice = np.array([
            [slice_offset + inner_indices_start] + [i for i in slice_origin],
            [inner_indices_stop - inner_indices_start] + [i for i in slice_shape],
        ])

        result.append((slice_idx, compressed_slice, read_ranges))
    return result


frms6_get_read_ranges = make_get_read_ranges(
    read_ranges_tile_block=_frms6_read_ranges_tile_block
)


def _make_decode_frms6(binning):
    @numba.njit(inline='always')
    def _decode_frms6(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
        """
        Row-for-row decoding of frms6 data
        """
        inp_decoded = inp.reshape((-1,)).view(native_dtype)
        out_3d = out.reshape(out.shape[0], -1, shape[-1])

        num_rows_binned = ds_shape[-2] // binning
        origin_y_binned = origin[1] // binning

        # how many rows need to be *read* for this tile
        # if we are un-binning, we need to read only
        rows_in_tile = shape[1] // binning

        # broadcast the row we read to (start, stop) rows:
        start = (idx % rows_in_tile) * binning
        stop = start + binning

        depth = idx // rows_in_tile

        # print(depth, start, stop, rows_in_tile, origin_y_binned)

        if origin_y_binned + (idx % rows_in_tile) < num_rows_binned // 2:
            out_3d[depth, start:stop, :] = inp_decoded
        else:
            out_3d[depth, start:stop, :] = inp_decoded[::-1]
    return _decode_frms6


decode_frms6_b1 = _make_decode_frms6(1)
decode_frms6_b2 = _make_decode_frms6(2)
decode_frms6_b4 = _make_decode_frms6(4)


class FRMS6Decoder(Decoder):
    def __init__(self, binning):
        self._binning = binning

    def get_decode(self, native_dtype, read_dtype):
        return {
            1: decode_frms6_b1,
            2: decode_frms6_b2,
            4: decode_frms6_b4,
        }[self._binning]


class FRMS6FileSet(FileSet):
    def __init__(self, global_header, *args, **kwargs):
        self._global_header = global_header
        super().__init__(*args, **kwargs)

    def _clone(self, *args, **kwargs):
        return self.__class__(
            global_header=self._global_header,
            *args, **kwargs
        )

    def get_read_ranges(
        self, start_at_frame: int, stop_before_frame: int,
        dtype, tiling_scheme: TilingScheme, sync_offset: int = 0,
        roi: typing.Union[np.ndarray, None] = None,
    ):
        fileset_arr = self.get_as_arr()
        binning = self._global_header['readoutmode']['bin']
        roi_nonzero = None
        if roi is not None:
            roi_nonzero = flat_nonzero(roi)
        return frms6_get_read_ranges(
            start_at_frame=start_at_frame,
            stop_before_frame=stop_before_frame,
            roi_nonzero=roi_nonzero,
            depth=tiling_scheme.depth,
            slices_arr=tiling_scheme.slices_array,
            fileset_arr=fileset_arr,
            sig_shape=tuple(tiling_scheme.dataset_shape.sig),
            sync_offset=sync_offset,
            bpp=np.dtype(dtype).itemsize * 8,
            frame_header_bytes=self._frame_header_bytes,
            frame_footer_bytes=self._frame_footer_bytes,
            extra=(binning,),
        )


class FRMS6DataSet(DataSet):
    r"""
    Read PNDetector FRMS6 files. FRMS6 data sets consist of multiple .frms6 files and
    a .hdr file. The first .frms6 file (matching \*_000.frms6) contains dark frames, which
    are subtracted if `enable_offset_correction` is true.

    Parameters
    ----------

    path : string
        Path to one of the files of the FRMS6 dataset (either .hdr or .frms6)

    enable_offset_correction : boolean
        Subtract dark frames when reading data

    gain_map_path : string
        Path to a gain map to apply (.mat format)

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    Examples
    --------

    >>> ds = ctx.load("frms6", path='./path_to_file.hdr', ...)  # doctest: +SKIP
    """

    def __init__(self, path, enable_offset_correction=True, gain_map_path=None, dest_dtype=None,
                 nav_shape=None, sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._path = path
        self._gain_map_path = gain_map_path
        self._dark_frame = None
        self._enable_offset_correction = enable_offset_correction
        self._meta = None
        self._filenames = None
        self._hdr_info = None
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        if dest_dtype is not None:
            warnings.warn(
                "dest_dtype is now handled per `get_tiles` call, and ignored here",
                DeprecationWarning
            )

    @property
    def shape(self):
        return self._meta.shape

    def _do_initialize(self):
        self._filenames = get_filenames(self._path)
        self._hdr_info = self._read_hdr_info()
        self._headers = [
            _read_file_header(path)
            for path in self._files()
        ]
        header = self._headers[0]
        raw_frame_size = header['height'], header['width']
        # frms6 frames are folded in a specific way, this is the shape after unfolding:
        frame_size = 2 * header['height'], header['width'] // 2
        assert header['width'] % 2 == 0
        hdr = self._get_hdr_info()
        bin_factor = hdr['readoutmode']['bin']
        if bin_factor > 1:
            frame_size = (frame_size[0] * bin_factor, frame_size[1])

        preferred_dtype = np.dtype("<u2")

        self._image_count = int(hdr['signalframes'])
        if self._nav_shape is None:
            self._nav_shape = tuple(hdr['stemimagesize'])
        if self._sig_shape is None:
            self._sig_shape = frame_size
        elif int(prod(self._sig_shape)) != int(prod(frame_size)):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(frame_size))
            )
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()

        if self._enable_offset_correction:
            preferred_dtype = np.dtype("float32")
        self._meta = DataSetMeta(
            raw_dtype=np.dtype("<u2"),
            dtype=preferred_dtype,
            metadata={'raw_frame_size': raw_frame_size},
            shape=Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape)),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        self._dark_frame = self._get_dark_frame()
        self._gain_map = self._get_gain_map()
        self._total_filesize = sum(
            os.stat(path).st_size
            for path in self._files()
        )
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    @classmethod
    def detect_params(cls, path, executor):
        hdr_filename = "%s.hdr" % executor.run_function(_get_base_filename, path)
        try:
            hdr = executor.run_function(_read_dataset_hdr, hdr_filename)
            bin_factor = hdr['readoutmode']['bin']
            nav_shape = tuple(hdr['stemimagesize'])
            image_count = int(prod(nav_shape))
            sig_shape = executor.run_function(_get_sig_shape, path, bin_factor)
        except Exception:
            return False
        return {
            "parameters": {
                "path": path,
                "nav_shape": nav_shape,
                "sig_shape": sig_shape,
            },
            "info": {
                "image_count": image_count,
                "native_sig_shape": sig_shape,
            }
        }

    @classmethod
    def get_supported_extensions(cls):
        return {"frms6", "hdr"}

    @classmethod
    def get_msg_converter(cls):
        return FRMS6DatasetParams

    def get_diagnostics(self):
        global_header = self._get_hdr_info()
        return [
            {"name": "Offset correction available and enabled",
             "value": str(self._dark_frame is not None)},
        ] + [
            {"name": str(k),
             "value": str(v)}
            for k, v in global_header.items()
        ]

    def _read_hdr_info(self):
        hdr_filename = "%s.hdr" % _get_base_filename(self._path)
        return _read_dataset_hdr(hdr_filename)

    def _get_hdr_info(self):
        return self._hdr_info

    def _get_dark_frame(self):
        if not self._enable_offset_correction:
            return None

        header = self._headers[0]
        num_frames = _num_frames(header)
        shape = (num_frames,) + tuple(self.shape.sig)
        fs = self._get_fileset([self._headers[0]])
        sig_dims = len(self.shape.sig)
        part_slice = Slice(
            origin=tuple([0] * len(shape)),
            shape=Shape(shape, sig_dims=sig_dims)
        )
        p = FRMS6Partition(
            meta=self._meta,
            partition_slice=part_slice,
            fileset=fs,
            start_frame=0,
            num_frames=num_frames,
            header=self._headers[0],
            io_backend=self.get_io_backend(),
            decoder=self.get_decoder(),
            binning=self._get_binning(),
        )
        tileshape = Shape(
            (128, 8) + (self.shape[-1],),
            sig_dims=2,
        )
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=tileshape,
            dataset_shape=self.shape,
        )
        tiles = p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype=np.float32)
        out = np.zeros(self.shape.sig, dtype=np.float32)
        for tile in tiles:
            out[tile.tile_slice.get(sig_only=True)] += np.sum(tile.data, axis=0)
        return out / num_frames

    def _get_gain_map(self):
        if self._gain_map_path is None:
            return None
        _, ext = os.path.splitext(self._gain_map_path)
        if ext.lower() == '.mat':
            gain_mat = sio.loadmat(self._gain_map_path)
            return gain_mat['GainMap']
        elif ext.lower() == '.csv':
            with open(self._gain_map_path) as csv_f:
                csv_reader = csv.reader(csv_f, dialect=GainMapCSVDialect)
                csv_gain_data = list(csv_reader)
                csv_gain_nums = [[float(x) for x in row if x != ''] for row in csv_gain_data]
                return np.array(csv_gain_nums).T

    def _files(self):
        return self._filenames

    def check_valid(self):
        try:
            for header in self._headers:
                if not _header_valid(header):
                    raise DataSetException(
                        "error while checking validity of %s" % header['path']
                    )
            return True
        except OSError as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path": self._path,
            "enable_offset_correction": self._enable_offset_correction,
            "gain_map_path": self._gain_map_path,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    @property
    def dtype(self):
        return self._meta.dtype

    def _get_fileset(self, headers=None):
        files = []
        start_idx = 0
        if headers is None:
            headers = self._headers[1:]  # skip first file w/ dark frames...
        header = headers[0]
        native_dtype = np.dtype(np.uint16)
        frame_header_size = header['frame_header_size']
        file_header = header['header_size']
        for header in headers:
            path = header['path']
            sig_shape = (header['height'], header['width'])
            num_frames = _num_frames(header)
            files.append(
                File(
                    path=path,
                    start_idx=start_idx,
                    end_idx=start_idx + num_frames,
                    native_dtype=native_dtype,
                    sig_shape=sig_shape,
                    file_header=file_header,
                    frame_header=frame_header_size,
                )
            )
            start_idx += num_frames
        global_header = self._get_hdr_info()
        return FRMS6FileSet(
            files=files,
            global_header=global_header,
            frame_header_bytes=frame_header_size
        )

    def get_correction_data(self):
        return CorrectionSet(
            dark=self._dark_frame,
            gain=self._gain_map,
        )

    def _get_binning(self):
        return self._get_hdr_info()['readoutmode']['bin']

    def get_decoder(self) -> Decoder:
        return FRMS6Decoder(
            binning=self._get_binning(),
        )

    def get_base_shape(self, roi):
        return (1, self._get_binning(), self.shape.sig[-1])

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield FRMS6Partition(
                meta=self._meta,
                partition_slice=part_slice,
                fileset=fileset,
                start_frame=start,
                num_frames=stop - start,
                binning=self._get_binning(),
                header=self._headers[0],
                io_backend=self.get_io_backend(),
                decoder=self.get_decoder(),
            )


class FRMS6Partition(BasePartition):
    def __init__(self, binning, header, *args, **kwargs):
        self._header = header
        self._binning = binning
        super().__init__(*args, **kwargs)

    def validate_tiling_scheme(self, tiling_scheme):
        binning = self._binning
        a = len(tiling_scheme.shape) == 3
        b = tiling_scheme.shape[1] % binning == 0
        c = tiling_scheme.shape[2] == self.meta.shape.sig[1]
        if not (a and b and c):
            raise ValueError(
                "Invalid tiling scheme: needs to be divisible by binning (%d)" % binning
            )



from libertem.common.scheduler import Worker, WorkerSet, Scheduler
from libertem.common.executor import ResourceDef


class MockTask:
    def __init__(self, resources: ResourceDef):
        self._resources = resources

    def get_resources(self) -> ResourceDef:
        return self._resources


def test_workers_for_task():
    cpu_resources: ResourceDef = {
        "compute": 1,
        "ndarray": 1,
        "CPU": 1,
    }
    cupy_resources: ResourceDef = {
        "compute": 1,
        "ndarray": 1,
        "CUDA": 1,
    }
    cuda_resources: ResourceDef = {
        "compute": 1,
        "CUDA": 1,
    }
    ws_cupy = WorkerSet([
        Worker(host='127.0.0.1', name='w_cupy', resources=cupy_resources, nthreads=1),
    ])
    ws_cuda = WorkerSet([
        Worker(host='127.0.0.1', name='w_cuda', resources=cuda_resources, nthreads=1),
    ])
    ws_cuda = ws_cuda.extend(ws_cupy)
    ws_numpy = WorkerSet([
        Worker(host='127.0.0.1', name='w_numpy', resources=cpu_resources, nthreads=1),
    ])
    ws = WorkerSet([
        Worker(host='127.0.0.1', name='w_empty', resources={}, nthreads=1),
    ])
    ws = ws.extend(ws_cuda)
    ws = ws.extend(ws_numpy)
    scheduler = Scheduler(all_workers=ws)

    task_cuda = MockTask(
        resources={'compute': 1, 'CUDA': 1}
    )
    task_cupy = MockTask(
        resources={'compute': 1, 'CUDA': 1, 'ndarray': 1}
    )
    task_numpy = MockTask(
        resources={'compute': 1, 'CPU': 1, 'ndarray': 1}
    )

    cuda_workers = scheduler.workers_for_task(task=task_cuda)
    assert len(cuda_workers) == 2
    assert cuda_workers == ws_cuda

    cupy_workers = scheduler.workers_for_task(task=task_cupy)
    assert len(cupy_workers) == 1
    assert cupy_workers == cupy_workers

    numpy_workers = scheduler.workers_for_task(task=task_numpy)
    assert len(numpy_workers) == 1
    assert numpy_workers == numpy_workers


def test_request_too_much():
    cpu_resources: ResourceDef = {
        "compute": 1,
        "ndarray": 1,
        "CPU": 1,
    }
    ws = WorkerSet([
        Worker(host='127.0.0.1', name='w_numpy', resources=cpu_resources, nthreads=1),
    ])
    scheduler = Scheduler(all_workers=ws)
    # the task requests a too large amount of the 'compute' resource:
    task_numpy = MockTask(
        resources={'compute': 2, 'CPU': 1, 'ndarray': 1}
    )
    numpy_workers = scheduler.workers_for_task(task=task_numpy)
    assert len(numpy_workers) == 0


def test_invalid_compares():
    assert not WorkerSet([]) == []
    assert not Worker(host='127.0.0.1', name='w_empty', resources={}, nthreads=1) == []
